# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from collections.abc import Mapping
from dataclasses import InitVar
from functools import cache
from typing import Any, cast

from pydantic import Field
from pydantic.dataclasses import dataclass

from ikigai.client.datax import (
    AppDict,
    CustomFacetArgumentDict,
    CustomFacetDict,
    CustomFacetVersionDict,
    DatasetDict,
    DatasetLogDict,
    FacetSpecsDict,
    FlowDefinitionDict,
    FlowDict,
    FlowLogDict,
    FlowStatusReportDict,
    GetComponentsForProjectResponse,
    GetDatasetMultipartUploadUrlsResponse,
    InitializeDatasetDownloadResponse,
    ModelDict,
    ModelSpecDict,
    ModelVersionDict,
    RunVariablesRequest,
    ScheduleDict,
)
from ikigai.client.session import Session
from ikigai.typing import Directory, ModelType, NamedDirectoryDict
from ikigai.utils import AppAccessLevel, CustomFacetAccessLevel
from ikigai.utils.missing import MISSING, MissingType

logger = logging.getLogger("ikigai.client.api")


@dataclass
class AccessAPI:
    # Init only vars
    session: InitVar[Session]

    __session: Session = Field(init=False)

    def __post_init__(self, session: Session) -> None:
        self.__session = session

    def generate_rootkit_token(self, script: str) -> str:
        # This is a weird endpoint, technically it should be under ComponentAPI
        # but it is more about giving privileges to a user to run arbitrary code.
        resp = self.__session.post(
            path="/component/generate-rootkit-token",
            json={"script": script},
        ).json()
        return resp["token"]


@dataclass
class ComponentAPI:
    # Init only vars
    session: InitVar[Session]

    __session: Session = Field(init=False)

    def __post_init__(self, session: Session) -> None:
        self.__session = session

    def __hash__(self) -> int:
        # Enable the usage of @cache on specs related apis
        return hash(id(self))

    """
    App APIs
    """

    def create_app(
        self,
        name: str,
        description: str,
        directory: Directory | None,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        resp = self.__session.post(
            path="/component/create-project",
            json={
                "project": {
                    "name": name,
                    "description": description,
                    "directory": directory_dict,
                },
            },
        ).json()
        return resp["project_id"]

    def get_app(self, app_id: str) -> AppDict:
        app_dict = self.__session.get(
            path="/component/get-project", params={"project_id": app_id}
        ).json()["project"]

        return cast(AppDict, app_dict)

    def get_app_by_name(self, name: str) -> AppDict:
        app_dict = self.__session.get(
            path="/component/get-project", params={"name": name}
        ).json()["project"]

        return cast(AppDict, app_dict)

    def get_apps_for_user(
        self, directory_id: str | MissingType = MISSING
    ) -> list[AppDict]:
        params: dict[str, Any] = {"fetch_all": directory_id is MISSING}
        if directory_id is not MISSING:
            params["directory_id"] = directory_id

        response: dict = self.__session.get(
            path="/component/get-projects-for-user", params=params
        ).json()

        app_dicts = response["projects"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[AppDict], app_dicts)

    def get_components_for_app(self, app_id: str) -> GetComponentsForProjectResponse:
        resp = self.__session.get(
            path="/component/get-components-for-project",
            params={"project_id": app_id},
        ).json()["project_components"][app_id]

        return cast(GetComponentsForProjectResponse, resp)

    def edit_app(
        self,
        app_id: str,
        name: str | MissingType = MISSING,
        directory: Directory | MissingType = MISSING,
        description: str | MissingType = MISSING,
    ) -> str:
        app: dict[str, Any] = {"project_id": app_id}

        if name is not MISSING:
            app["name"] = name
        if directory is not MISSING:
            app["directory"] = directory.to_dict()
        if description is not MISSING:
            app["description"] = description

        resp = self.__session.post(
            path="/component/edit-project",
            json={"project": app},
        ).json()

        return resp["project_id"]

    def delete_app(self, app_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-project",
            json={"project": {"project_id": app_id}},
        ).json()

        return resp["project_id"]

    def grant_app_access(
        self, app_id: str, email: str, access_level: AppAccessLevel
    ) -> str:
        resp = self.__session.post(
            path="/component/share-project",
            json={
                "project": {"project_id": app_id},
                "user": {"email": email, "project_access_level": access_level},
            },
        ).json()
        return resp["project_id"]

    def update_app_access(
        self, app_id: str, email: str, access_level: AppAccessLevel
    ) -> str:
        resp = self.__session.post(
            path="/component/edit-project-access-level-for-user",
            json={
                "project": {
                    "project_id": app_id,
                },
                "user": {"email": email, "project_access_level": access_level},
            },
        ).json()
        return resp["project_id"]

    def revoke_app_access(self, app_id: str, email: str) -> str:
        resp = self.__session.post(
            path="/component/unshare-project",
            json={
                "project": {"project_id": app_id},
                "user": {"email": email},
            },
        ).json()
        return resp["project_id"]

    """
    Custom Facet APIs
    """

    def create_custom_facet(
        self,
        name: str,
        chain_group: str,
        description: str,
        tags: list[str],
        python_script: str,
        libraries: list[str],
        rootkit_token: str,
        arguments: list[CustomFacetArgumentDict],
    ) -> str:
        resp = self.__session.post(
            path="/component/create-custom-facet",
            json={
                "custom_facet": {
                    "name": name,
                    "chain_group": chain_group,
                    "description": description,
                    "tags": tags,
                    "python_script": python_script,
                    "libraries": libraries,
                    "rootkit_token": rootkit_token,
                    "arguments": arguments,
                }
            },
        ).json()
        return resp["custom_facet_id"]

    def get_custom_facet(self, custom_facet_id: str) -> CustomFacetDict:
        custom_facet_dict = self.__session.get(
            path="/component/get-custom-facet",
            params={"custom_facet_id": custom_facet_id},
        ).json()["custom_facet"]

        return cast(CustomFacetDict, custom_facet_dict)

    def get_custom_facet_by_name(self, name: str) -> CustomFacetDict:
        custom_facet_dict = self.__session.get(
            path="/component/get-custom-facet", params={"name": name}
        ).json()["custom_facet"]

        return cast(CustomFacetDict, custom_facet_dict)

    def get_custom_facets_for_user(self) -> list[CustomFacetDict]:
        custom_facet_dicts = self.__session.get(
            path="/component/get-custom-facets-for-user",
        ).json()["custom_facets"]

        return cast(list[CustomFacetDict], custom_facet_dicts)

    def edit_custom_facet(
        self,
        custom_facet_id: str,
        chain_group: str,
        name: str | MissingType = MISSING,
        description: str | MissingType = MISSING,
        tags: list[str] | MissingType = MISSING,
        python_script: str | MissingType = MISSING,
        libraries: list[str] | MissingType = MISSING,
        rootkit_token: str | MissingType = MISSING,
        arguments: list[CustomFacetArgumentDict] | MissingType = MISSING,
    ) -> str:
        custom_facet: dict[str, Any] = {
            "custom_facet_id": custom_facet_id,
            "chain_group": chain_group,
        }

        if name is not MISSING:
            custom_facet["name"] = name
        if description is not MISSING:
            custom_facet["description"] = description
        if tags is not MISSING:
            custom_facet["tags"] = tags
        if python_script is not MISSING:
            custom_facet["python_script"] = python_script
        if libraries is not MISSING:
            custom_facet["libraries"] = libraries
        if rootkit_token is not MISSING:
            custom_facet["rootkit_token"] = rootkit_token
        if arguments is not MISSING:
            custom_facet["arguments"] = arguments

        resp = self.__session.post(
            path="/component/edit-custom-facet",
            json={
                "custom_facet": custom_facet,
                "save_as_version": False,
            },
        ).json()

        return resp["custom_facet_id"]

    def delete_custom_facet(self, custom_facet_id: str) -> None:
        self.__session.post(
            path="/component/delete-custom-facet",
            json={"custom_facet": {"custom_facet_id": custom_facet_id}},
        ).json()
        return None

    """
    Custom Facet Version APIs
    """

    def get_custom_facet_version(
        self, custom_facet_id: str, version_id: str
    ) -> CustomFacetVersionDict:
        resp = self.__session.get(
            path="/component/get-version-for-custom-facet",
            params={"custom_facet_id": custom_facet_id, "version_id": version_id},
        ).json()
        return cast(CustomFacetVersionDict, resp["custom_facet_version"])

    def get_custom_facet_versions(
        self, custom_facet_id: str
    ) -> list[CustomFacetVersionDict]:
        response = self.__session.get(
            path="/component/get-versions-for-custom-facet",
            params={"custom_facet_id": custom_facet_id},
        ).json()
        if warning := response["limit_warning"]:
            logger.warning(warning)

        custom_facet_version_dicts = response["custom_facet_versions"]
        return cast(list[CustomFacetVersionDict], custom_facet_version_dicts)

    def create_custom_facet_version(
        self,
        custom_facet_id: str,
        version: str,
        description: str,
        python_script: str,
        libraries: list[str],
        rootkit_token: str,
        arguments: list[CustomFacetArgumentDict],
    ) -> str:
        response = self.__session.post(
            path="/component/edit-custom-facet",
            json={
                "save_as_version": True,
                "custom_facet_version": {
                    "custom_facet_id": custom_facet_id,
                    "version": version,
                    "python_script": python_script,
                    "libraries": libraries,
                    "rootkit_token": rootkit_token,
                    "arguments": arguments,
                    "description": description,
                },
            },
        ).json()

        if warning := response["limit_warning"]:
            logger.warning(warning)

        return response["version_id"]

    def grant_custom_facet_access(
        self, custom_facet_id: str, email: str, access_level: CustomFacetAccessLevel
    ) -> str:
        resp = self.__session.post(
            path="/component/share-custom-facet",
            json={
                "custom_facet": {"custom_facet_id": custom_facet_id},
                "user": {"email": email},
                "access_level": access_level,
            },
        ).json()
        return cast(str, resp["custom_facet_id"])

    def update_custom_facet_access(
        self, custom_facet_id: str, email: str, access_level: CustomFacetAccessLevel
    ) -> str:
        resp = self.__session.post(
            path="/component/edit-custom-facet-access-level",
            json={
                "custom_facet": {"custom_facet_id": custom_facet_id},
                "user": {"email": email},
                "access_level": access_level,
            },
        ).json()
        return cast(str, resp["custom_facet_id"])

    def revoke_custom_facet_access(self, custom_facet_id: str, email: str) -> str:
        resp = self.__session.post(
            path="/component/edit-custom-facet-access-level",
            json={
                "custom_facet": {"custom_facet_id": custom_facet_id},
                "user": {"email": email},
                "access_level": "NO_ACCESS",
            },
        ).json()
        return cast(str, resp["custom_facet_id"])

    """
    Dataset APIs
    """

    def create_dataset(
        self, app_id: str, name: str, directory: Directory | None
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        resp = self.__session.post(
            path="/component/create-dataset",
            json={
                "dataset": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                },
            },
        ).json()
        return resp["dataset_id"]

    def initialize_dataset_download(
        self, app_id: str, dataset_id: str
    ) -> InitializeDatasetDownloadResponse:
        resp = self.__session.get(
            path="/component/get-dataset-download-url",
            params={
                "project_id": app_id,
                "dataset_id": dataset_id,
            },
        ).json()
        return cast(InitializeDatasetDownloadResponse, resp)

    def get_dataset(self, app_id: str, dataset_id: str) -> DatasetDict:
        resp = self.__session.get(
            path="/component/get-dataset",
            params={"project_id": app_id, "dataset_id": dataset_id},
        ).json()
        dataset = resp["dataset"]

        return cast(DatasetDict, dataset)

    def get_dataset_by_name(self, app_id: str, name: str) -> DatasetDict:
        resp = self.__session.get(
            path="/component/get-dataset",
            params={"project_id": app_id, "name": name},
        ).json()
        dataset = resp["dataset"]

        return cast(DatasetDict, dataset)

    def get_datasets_for_app(
        self, app_id: str, directory_id: str | MissingType = MISSING
    ) -> list[DatasetDict]:
        params = {"project_id": app_id, "fetch_all": directory_id is MISSING}
        if directory_id is not MISSING:
            params["directory_id"] = directory_id

        response = self.__session.get(
            path="/component/get-datasets-for-project",
            params=params,
        ).json()
        datasets = response["datasets"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[DatasetDict], datasets)

    def get_dataset_multipart_upload_urls(
        self, dataset_id: str, app_id: str, filename: str, file_size: int
    ) -> GetDatasetMultipartUploadUrlsResponse:
        resp = self.__session.get(
            path="/component/get-dataset-multipart-upload-urls",
            params={
                "dataset_id": dataset_id,
                "project_id": app_id,
                "filename": filename,
                "file_size": file_size,
            },
        ).json()

        return GetDatasetMultipartUploadUrlsResponse(
            upload_id=resp["upload_id"],
            content_type=resp["content_type"],
            urls={
                int(chunk_idx): upload_url
                for chunk_idx, upload_url in resp["urls"].items()
            },
        )

    def get_dataset_log(
        self, app_id: str, dataset_id: str, limit: int = 5
    ) -> list[DatasetLogDict]:
        dataset_log = self.__session.get(
            path="/component/get-dataset-log",
            params={"dataset_id": dataset_id, "project_id": app_id, "limit": limit},
        ).json()["dataset_log"]

        return cast(list[DatasetLogDict], dataset_log)

    def edit_dataset(
        self,
        app_id: str,
        dataset_id: str,
        name: str | MissingType = MISSING,
        directory: Directory | MissingType = MISSING,
    ) -> str:
        dataset: dict[str, Any] = {
            "project_id": app_id,
            "dataset_id": dataset_id,
        }

        if name is not MISSING:
            dataset["name"] = name
        if directory is not MISSING:
            dataset["directory"] = directory.to_dict()

        resp = self.__session.post(
            path="/component/edit-dataset",
            json={
                "dataset": dataset,
            },
        ).json()

        return resp["dataset_id"]

    def verify_dataset_upload(
        self, app_id: str, dataset_id: str, filename: str
    ) -> None:
        self.__session.get(
            path="/component/verify-dataset-upload",
            params={
                "project_id": app_id,
                "dataset_id": dataset_id,
                "filename": filename,
            },
        )
        return None

    def confirm_dataset_upload(self, app_id: str, dataset_id: str) -> str:
        resp = self.__session.get(
            path="/component/confirm-dataset-upload",
            params={"project_id": app_id, "dataset_id": dataset_id},
        ).json()
        return resp["status"]

    def abort_datset_multipart_upload(
        self, app_id: str, dataset_id: str, filename: str, upload_id: str
    ) -> None:
        self.__session.post(
            path="/component/complete-dataset-multipart-upload",
            json={
                "abort": True,
                "dataset": {
                    "dataset_id": dataset_id,
                    "project_id": app_id,
                    "filename": filename,
                },
                "upload_id": upload_id,
            },
        )
        return None

    def complete_datset_multipart_upload(
        self,
        app_id: str,
        dataset_id: str,
        filename: str,
        upload_id: str,
        etags: dict[int, str],
    ) -> None:
        self.__session.post(
            path="/component/complete-dataset-multipart-upload",
            json={
                "abort": False,
                "dataset": {
                    "dataset_id": dataset_id,
                    "project_id": app_id,
                    "filename": filename,
                },
                "upload_id": upload_id,
                "etags": etags,
            },
        )
        return None

    def delete_dataset(self, app_id: str, dataset_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-dataset",
            json={"dataset": {"project_id": app_id, "dataset_id": dataset_id}},
        ).json()

        return resp["dataset_id"]

    """
    Flow APIs
    """

    def create_flow(
        self,
        app_id: str,
        name: str,
        directory: Directory | None,
        high_volume_preference: bool,
        flow_definition: FlowDefinitionDict,
        schedule: ScheduleDict | None,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )
        schedule_dict = schedule if schedule else None
        resp = self.__session.post(
            path="/component/create-pipeline",
            json={
                "pipeline": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                    "high_volume_preference": high_volume_preference,
                    "definition": flow_definition,
                    "schedule": schedule_dict,
                },
            },
        ).json()
        return resp["pipeline_id"]

    def get_flow(self, flow_id: str) -> FlowDict:
        flow = self.__session.get(
            path="/component/get-pipeline", params={"pipeline_id": flow_id}
        ).json()["pipeline"]

        return cast(FlowDict, flow)

    def get_flow_by_name(self, app_id: str, name: str) -> FlowDict:
        flow = self.__session.get(
            path="/component/get-pipeline", params={"project_id": app_id, "name": name}
        ).json()["pipeline"]

        return cast(FlowDict, flow)

    def get_flows_for_app(
        self, app_id: str, directory_id: str | MissingType = MISSING
    ) -> list[FlowDict]:
        params = {"project_id": app_id, "fetch_all": directory_id is MISSING}
        if directory_id is not MISSING:
            params["directory_id"] = directory_id

        response = self.__session.get(
            path="/component/get-pipelines-for-project",
            params=params,
        ).json()

        flows = response["pipelines"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[FlowDict], flows)

    def get_flow_log(
        self, app_id: str, flow_id: str, max_count: int
    ) -> list[FlowLogDict]:
        log_dicts = self.__session.get(
            path="/component/get-pipeline-log",
            params={
                "pipeline_id": flow_id,
                "project_id": app_id,
                "limit": max_count,
            },
        ).json()["pipeline_log"]

        return cast(list[FlowLogDict], log_dicts)

    def edit_flow(
        self,
        app_id: str,
        flow_id: str,
        name: str | None = None,
        directory: Directory | None = None,
        high_volume_preference: bool | None = None,
        flow_definition: FlowDefinitionDict | None = None,
        schedule: ScheduleDict | None | MissingType = MISSING,
    ) -> str:
        pipeline: dict[str, Any] = {
            "project_id": app_id,
            "pipeline_id": flow_id,
        }

        if name is not None:
            pipeline["name"] = name
        if directory is not None:
            pipeline["directory"] = directory.to_dict()
        if high_volume_preference is not None:
            pipeline["high_volume_preference"] = high_volume_preference
        if flow_definition is not None:
            pipeline["definition"] = flow_definition
        if schedule is not MISSING:
            if schedule is not None:
                pipeline["schedule"] = schedule
            else:
                # HACK: No way to remove schedule via API, update after BE supports it.
                pipeline["schedule"] = {
                    "name": "-",
                    "cron": "0 0 0 0 0",
                    "start_time": "1",
                    "end_time": "1",
                }

        resp = self.__session.post(
            path="/component/edit-pipeline", json={"pipeline": pipeline}
        ).json()
        return resp["pipeline_id"]

    def delete_flow(self, app_id: str, flow_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-pipeline",
            json={"pipeline": {"project_id": app_id, "pipeline_id": flow_id}},
        ).json()

        return resp["pipeline_id"]

    def run_flow(
        self, app_id: str, flow_id: str, variables: RunVariablesRequest
    ) -> str:
        payload: dict[str, Mapping] = {
            "pipeline": {"project_id": app_id, "pipeline_id": flow_id}
        }
        if variables:
            payload["variables"] = variables

        resp = self.__session.post(
            path="/component/run-pipeline",
            json=payload,
        ).json()

        return resp["pipeline_id"]

    def is_flow_runing(self, app_id: str, flow_id: str) -> FlowStatusReportDict:
        resp = self.__session.get(
            path="/component/is-pipeline-running",
            params={"project_id": app_id, "pipeline_id": flow_id},
        ).json()

        # BE is a bit inconsistent with the response so clean it up
        status = resp["progress"]["status"] if resp["status"] else "IDLE"
        progress = resp["progress"].get("progress")
        message = resp["progress"].get("message")

        return FlowStatusReportDict(
            status=status,
            progress=progress,
            message=message,
        )

    """
    Model APIs
    """

    def create_model(
        self,
        app_id: str,
        name: str,
        directory: Directory | None,
        model_type: ModelType,
        description: str,
    ) -> str:
        directory_dict = (
            cast(dict, directory.to_dict()) if directory is not None else {}
        )

        resp = self.__session.post(
            path="/component/create-model",
            json={
                "model": {
                    "project_id": app_id,
                    "name": name,
                    "directory": directory_dict,
                    "model_type": model_type.model_type,
                    "sub_model_type": model_type.sub_model_type,
                    "description": description,
                }
            },
        ).json()

        return resp["model_id"]

    def get_model(self, app_id: str, model_id: str) -> ModelDict:
        model = self.__session.get(
            path="/component/get-model",
            params={"project_id": app_id, "model_id": model_id},
        ).json()["model"]

        return cast(ModelDict, model)

    def get_model_by_name(self, app_id: str, name: str) -> ModelDict:
        model = self.__session.get(
            path="/component/get-model",
            params={"project_id": app_id, "name": name},
        ).json()["model"]

        return cast(ModelDict, model)

    def get_models_for_app(
        self, app_id: str, directory_id: str | MissingType = MISSING
    ) -> list[ModelDict]:
        params = {"project_id": app_id, "fetch_all": directory_id is MISSING}
        if directory_id is not MISSING:
            params["directory_id"] = directory_id

        response = self.__session.get(
            path="/component/get-models-for-project",
            params=params,
        ).json()

        models = response["models"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[ModelDict], models)

    def edit_model(
        self,
        app_id: str,
        model_id: str,
        name: str | MissingType = MISSING,
        directory: Directory | MissingType = MISSING,
        description: str | MissingType = MISSING,
    ) -> str:
        model: dict[str, Any] = {
            "project_id": app_id,
            "model_id": model_id,
        }

        if name is not MISSING:
            model["name"] = name
        if directory is not MISSING:
            model["directory"] = directory.to_dict()
        if description is not MISSING:
            model["description"] = description

        resp = self.__session.post(
            path="/component/edit-model",
            json={"model": model},
        ).json()

        return resp["model_id"]

    def delete_model(self, app_id: str, model_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-model",
            json={"model": {"project_id": app_id, "model_id": model_id}},
        ).json()

        return resp["model_id"]

    """
    Model Version APIs
    """

    def get_model_version(self, app_id: str, version_id: str) -> ModelVersionDict:
        resp = self.__session.get(
            path="/component/get-model-version",
            params={"project_id": app_id, "version_id": version_id},
        ).json()
        model_version = resp["model_version"]

        return cast(ModelVersionDict, model_version)

    def get_model_versions(self, app_id: str, model_id: str) -> list[ModelVersionDict]:
        resp = self.__session.get(
            path="/component/get-model-versions",
            params={"project_id": app_id, "model_id": model_id},
        ).json()
        model_versions = resp["versions"]

        return cast(list[ModelVersionDict], model_versions)

    """
    Directory APIs
    """

    def create_app_directory(self, name: str, parent: Directory | None = None) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-project-directory",
            json={
                "directory": {
                    "name": name,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_app_directory(self, directory_id: str) -> NamedDirectoryDict:
        directory = self.__session.get(
            path="/component/get-project-directory",
            params={"directory_id": directory_id},
        ).json()["directory"]

        return cast(NamedDirectoryDict, directory)

    def get_app_directories_for_user(
        self, directory_id: str | MissingType = MISSING
    ) -> list[NamedDirectoryDict]:
        """
        NOTE: Platform should add a fetch_all field to distinguish between root
            directory and no directory, in anticipation of this we just pass it in.
        """
        params: dict[str, Any] = {"fetch_all": directory_id is MISSING}
        if directory_id is not MISSING:
            params["directory_id"] = directory_id

        directory_dicts = self.__session.get(
            path="/component/get-project-directories-for-user",
            params=params,
        ).json()["directories"]

        return cast(list[NamedDirectoryDict], directory_dicts)

    def delete_app_directory(self, directory_id: str) -> str:
        resp = self.__session.post(
            path="/component/delete-project-directory",
            json={"directory": {"directory_id": directory_id}},
        ).json()

        return resp["directory_id"]

    def create_dataset_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-dataset-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_dataset_directory(
        self, app_id: str, directory_id: str
    ) -> NamedDirectoryDict:
        directory = self.__session.get(
            path="/component/get-dataset-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(NamedDirectoryDict, directory)

    def get_dataset_directories_for_app(
        self, app_id: str, parent: Directory | MissingType = MISSING
    ) -> list[NamedDirectoryDict]:
        params = {"project_id": app_id}
        if parent is not MISSING:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-dataset-directories-for-project",
            params=params,
        ).json()
        directories = resp["directories"]

        return cast(list[NamedDirectoryDict], directories)

    def create_flow_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-pipeline-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_flow_directory(self, app_id: str, directory_id: str) -> NamedDirectoryDict:
        directory = self.__session.get(
            path="/component/get-pipeline-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(NamedDirectoryDict, directory)

    def get_flow_directories_for_app(
        self, app_id: str, parent: Directory | MissingType = MISSING
    ) -> list[NamedDirectoryDict]:
        params = {"project_id": app_id}
        if parent is not MISSING:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-pipeline-directories-for-project",
            params=params,
        ).json()

        directories = resp["directories"]
        return cast(list[NamedDirectoryDict], directories)

    def create_model_directory(
        self, app_id: str, name: str, parent: Directory | None = None
    ) -> str:
        parent_id = parent.directory_id if parent else ""

        resp = self.__session.post(
            path="/component/create-model-directory",
            json={
                "directory": {
                    "name": name,
                    "project_id": app_id,
                    "parent_id": parent_id,
                }
            },
        ).json()

        return resp["directory_id"]

    def get_model_directory(self, app_id: str, directory_id: str) -> NamedDirectoryDict:
        directory = self.__session.get(
            path="/component/get-model-directory",
            params={"project_id": app_id, "directory_id": directory_id},
        ).json()["directory"]

        return cast(NamedDirectoryDict, directory)

    def get_model_directories_for_app(
        self, app_id: str, parent: Directory | MissingType = MISSING
    ) -> list[NamedDirectoryDict]:
        params = {"project_id": app_id}
        if parent is not MISSING:
            params["directory_id"] = parent.directory_id

        resp = self.__session.get(
            path="/component/get-model-directories-for-project",
            params=params,
        ).json()

        directories = resp["directories"]
        return cast(list[NamedDirectoryDict], directories)

    """
    Spec APIs
    """

    @cache
    def get_facet_specs(self) -> FacetSpecsDict:
        resp = self.__session.get(
            path="/component/get-facet-specs",
        ).json()

        return cast(FacetSpecsDict, resp)

    @cache
    def get_model_specs(self) -> list[ModelSpecDict]:
        resp = self.__session.get(
            path="/component/get-model-specs",
        ).json()

        model_specs = resp.values()

        return cast(list[ModelSpecDict], model_specs)


@dataclass
class SearchAPI:
    # Init only vars
    session: InitVar[Session]

    __session: Session = Field(init=False)

    def __post_init__(self, session: Session) -> None:
        self.__session = session

    def heartbeat(self) -> None:
        self.__session.get(path="/search/heartbeat", suppress_logging=True)
        return None

    """
    Search APIs
    """

    def search_projects_for_user(self, query: str) -> list[AppDict]:
        response = self.__session.get(
            path="/search/search-projects-for-user", params={"query": query}
        ).json()

        app_dicts = response["projects"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[AppDict], app_dicts)

    def search_custom_facets_for_user(self, query: str) -> list[CustomFacetDict]:
        response = self.__session.get(
            path="/search/search-custom-facets-for-user",
            params={"query": query},
        ).json()

        custom_facet_dicts = response["custom_facets"]
        return cast(list[CustomFacetDict], custom_facet_dicts)

    def search_datasets_for_project(self, app_id: str, query: str) -> list[DatasetDict]:
        response = self.__session.get(
            path="/search/search-datasets-for-project",
            params={"project_id": app_id, "query": query},
        ).json()

        dataset_dict = response["datasets"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[DatasetDict], dataset_dict)

    def search_flows_for_project(self, app_id: str, query: str) -> list[FlowDict]:
        response = self.__session.get(
            path="/search/search-pipelines-for-project",
            params={"project_id": app_id, "query": query},
        ).json()

        flow_dict = response["pipelines"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[FlowDict], flow_dict)

    def search_models_for_project(self, app_id: str, query: str) -> list[ModelDict]:
        response = self.__session.get(
            path="/search/search-models-for-project",
            params={"project_id": app_id, "query": query},
        ).json()

        model_dict = response["models"]
        if warning := response["limit_warning"]:
            logger.warning(warning)

        return cast(list[ModelDict], model_dict)



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from collections.abc import Generator, Mapping
from functools import cached_property
from typing import Any

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator,
    model_validator,
)

from ikigai.client import datax
from ikigai.typing import Helpful
from ikigai.typing.pydantic_extensions import LowercaseStr
from ikigai.utils import (
    ModelHyperparameterType,
    ModelParameterType,
)
from ikigai.utils.compatibility import Self, override
from ikigai.utils.missing import MISSING, MissingType

logger = logging.getLogger("ikigai.components.specs")


class ModelMetricsSpec(RootModel, Helpful):
    root: dict[LowercaseStr, Any]

    @model_validator(mode="after")
    def validate_lowercase_keys(self) -> Self:
        self.root = {metric.lower(): value for metric, value in self.root.items()}
        return self

    def __contains__(self, name: str) -> bool:
        return name.lower() in self.root

    def __getitem__(self, name: str) -> Any:
        if name not in self:
            error_msg = f"{name.title()} metric does not exist"
            raise AttributeError(error_msg)
        return self.root[name.lower()]

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Self:
        return cls.model_validate({key.lower(): value for key, value in data.items()})

    @override
    def _help(self) -> Generator[str]:
        yield "metrics:"
        if not self.root:
            yield "  No metrics"
            return

        for metric, value in self.root.items():
            yield f"  {metric}: {value}"


class ModelParameterSpec(BaseModel, Helpful):
    name: str
    default_value: Any | None = None
    have_options: bool
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    options: list | None = None
    parameter_type: str

    model_config = ConfigDict(frozen=True)

    def __validation_error_message(
        self, model: str, expectation: str, actuals: MissingType | Any = MISSING
    ) -> str:
        if actuals is MISSING:
            actuals_str = ""
        elif actuals is None:
            actuals_str = ", got 'None'"
        elif isinstance(actuals, type):
            actuals_str = f", got type '{actuals.__name__}'"
        else:
            actuals_str = f", got {actuals.__class__.__name__}({actuals!r})"

        return f"Parameter '{self.name}' for {model} {expectation}{actuals_str}"

    def validate_value(self, model: str, value: Any) -> None:
        if self.is_list:
            return self.__validate_list_value(model, value)

        # Not a list or dict, so it must be a scalar value
        return self.__validate_scalar_value(model, value)

    def __validate_list_value(self, model: str, value: Any) -> None:
        if not isinstance(value, list):
            error_msg = self.__validation_error_message(model, "must be list", value)
            raise TypeError(error_msg)

        scalar_parameter_spec = self.model_copy(update={"is_list": False})
        for item in value:
            scalar_parameter_spec.validate_value(model, item)
        return None  # All items validated

    def __validate_scalar_value(self, model: str, value: Any) -> None:
        if self.options and value not in self.options:
            error_msg = self.__validation_error_message(
                model, "must be one of {self.options}", value
            )
            raise ValueError(error_msg)

        if self.parameter_type == ModelParameterType.BOOLEAN and not isinstance(
            value, bool
        ):
            error_msg = self.__validation_error_message(model, "must be boolean", value)
            raise TypeError(error_msg)

        if self.parameter_type == ModelParameterType.TEXT and not isinstance(
            value, str
        ):
            error_msg = self.__validation_error_message(model, "must be string", value)
            raise TypeError(error_msg)

        if self.parameter_type == ModelParameterType.NUMBER and not isinstance(
            value, int | float
        ):
            error_msg = self.__validation_error_message(model, "must be numeric", value)
            raise TypeError(error_msg)

    @override
    def _help(self) -> Generator[str]:
        parameter_type = (
            f"{self.parameter_type}"
            if not self.is_list
            else f"list[{self.parameter_type}]"
        )
        parameter_value = (
            f" = {self.default_value!r}" if self.default_value is not None else ""
        ) + (
            f"  options=[{'|'.join([str(option) for option in self.options])}]"
            if self.options
            else ""
        )
        yield f"{self.name}: {parameter_type}{parameter_value}"


class ModelHyperparameterSpec(BaseModel, Helpful):
    name: str
    default_value: Any | None = None
    have_options: bool
    have_sub_hyperparameters: bool
    hyperparameter_group: str | None
    hyperparameter_type: ModelHyperparameterType
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    children: dict[str, ModelHyperparameterSpec]
    options: list | None = None
    sub_hyperparameter_requirements: dict[Any, list[str]] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)

    @field_validator("sub_hyperparameter_requirements", mode="before")
    @classmethod
    def validate_sub_hyperparameter_requirements(cls, v: Any) -> dict[Any, list[str]]:
        if isinstance(v, list):
            if not (
                all(
                    isinstance(item, tuple | list) and len(item) == 2  # noqa: PLR2004 -- 2 is a fairly reasonable literal len
                    for item in v
                )
            ):
                error_msg = (
                    "Expected a list of "
                    "(hyperparameter_value, [required_hyperparameter_names]) tuples"
                )
                raise ValueError(error_msg)
            return dict(v)

        if isinstance(v, dict):
            return v

        error_msg = "Expected a dictionary or list of requirements"
        raise ValueError(error_msg)

    def __validation_error_message(
        self, model: str, expectation: str, actuals: MissingType | Any = MISSING
    ) -> str:
        if actuals is MISSING:
            actuals_str = ""
        elif actuals is None:
            actuals_str = ", got 'None'"
        elif isinstance(actuals, type):
            actuals_str = f", got type '{actuals.__name__}'"
        else:
            actuals_str = f", got {actuals.__class__.__name__}({actuals!r})"

        return f"Hyperparameter '{self.name}' for {model} {expectation}{actuals_str}"

    def validate_value(self, model: str, value: Any) -> None:
        if self.is_list:
            return self.__validate_list_value(model, value)

        if self.hyperparameter_type == ModelHyperparameterType.MAP:
            return self.__validate_dict_value(model, value)

        # Not a list or dict, so it must be a scalar value
        return self.__validate_scalar_value(model, value)

    def __validate_list_value(self, model: str, value: Any) -> None:
        if not isinstance(value, list):
            error_msg = self.__validation_error_message(model, "must be list", value)
            raise TypeError(error_msg)

        scalar_hyperparameter_spec = self.model_copy(update={"is_list": False})
        for item in value:
            scalar_hyperparameter_spec.validate_value(model, item)
        return None  # All items validated

    def __validate_dict_value(self, model: str, value: Any) -> None:
        if not isinstance(value, Mapping):
            error_msg = self.__validation_error_message(model, "must be mapping", value)
            raise TypeError(error_msg)

        for name, child_value in value.items():
            if name not in self.children:
                error_msg = self.__validation_error_message(
                    model, f"provided with unexpected child hyperparameter '{name}'"
                )
                raise KeyError(error_msg)
            child_spec = self.children[name]
            child_spec.validate_value(model, child_value)
        return None  # All child hyperparameters validated

    def __validate_scalar_value(self, model: str, value: Any) -> None:
        if self.options and value not in self.options:
            error_msg = self.__validation_error_message(
                model, f"must be one of {self.options}", value
            )
            raise ValueError(error_msg)

        if (
            self.hyperparameter_type == ModelHyperparameterType.BOOLEAN
            and not isinstance(value, bool)
        ):
            error_msg = self.__validation_error_message(model, "must be boolean", value)
            raise TypeError(error_msg)

        if self.hyperparameter_type == ModelHyperparameterType.TEXT and not isinstance(
            value, str
        ):
            error_msg = self.__validation_error_message(model, "must be string", value)
            raise TypeError(error_msg)

        if (
            self.hyperparameter_type == ModelHyperparameterType.NUMBER
            and not isinstance(value, int | float)
        ):
            error_msg = self.__validation_error_message(model, "must be numeric", value)
            raise TypeError(error_msg)

    @override
    def _help(self) -> Generator[str]:
        hyperparameter_type = (
            f"{self.hyperparameter_type}"
            if not self.is_list
            else f"list[{self.hyperparameter_type}]"
        )
        hyperparameter_value = (
            f" = {self.default_value!r}" if self.default_value is not None else ""
        ) + (
            f"  options=[{'|'.join([str(option) for option in self.options])}]"
            if self.options
            else ""
        )
        if not self.children:
            yield f"{self.name}: {hyperparameter_type}{hyperparameter_value}"

        if self.children:
            yield f"{self.name}: {hyperparameter_type} = {{"
            for child in self.children.values():
                if child.is_hidden:
                    continue
                yield from (f"  {child_help}" for child_help in child._help())
            yield "}"


class SubModelSpec(BaseModel, Helpful):
    name: str
    model_type: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    metrics: ModelMetricsSpec
    parameters: dict[str, ModelParameterSpec]
    hyperparameters: dict[str, ModelHyperparameterSpec]

    model_config = ConfigDict(frozen=True)

    @field_validator("parameters", mode="after")
    @classmethod
    def validate_parameters(
        cls, v: dict[str, ModelParameterSpec]
    ) -> dict[str, ModelParameterSpec]:
        return {parameter.name: parameter for parameter in v.values()}

    @field_validator("hyperparameters", mode="after")
    @classmethod
    def validate_hyperparameters(
        cls, v: dict[str, ModelHyperparameterSpec]
    ) -> dict[str, ModelHyperparameterSpec]:
        # If one hyperparameter has a group, then all hyperparameters must have groups
        has_any_groups = any(
            hyperparameter.hyperparameter_group is not None
            for hyperparameter in v.values()
        )
        has_all_groups = all(
            hyperparameter.hyperparameter_group is not None
            for hyperparameter in v.values()
        )

        if has_any_groups and not has_all_groups:
            message = (
                "Inconsistent hyperparameter groups for: "
                "Some hyperparameters have groups while others do not.\n"
                "This is likely a due to a bug in the model specification."
            )
            logger.error(message, extra={"hyperparameter specification": v})
            raise ValueError(message)
        return {hyperparameter.name: hyperparameter for hyperparameter in v.values()}

    @property
    def sub_model_type(self) -> str:
        return self.name

    @classmethod
    def from_dict(cls, model_type: str, data: datax.SubModelSpecDict) -> Self:
        data_dict = {
            **data,
            "model_type": model_type,
        }

        return cls.model_validate(data_dict)

    @cached_property
    def _hyperparameter_groups(
        self,
    ) -> dict[datax.HyperParameterName, datax.HyperParameterGroupName]:
        # Create a mapping from hyperparameter name to its group
        return {
            hyperparameter.name: hyperparameter.hyperparameter_group
            for hyperparameter in self.hyperparameters.values()
            if hyperparameter.hyperparameter_group
        }

    @override
    def _help(self) -> Generator[str]:
        # Sub-model name
        yield f"{self.name.title()}:"
        if self.keywords:
            yield f"  keywords: {self.keywords}"

        yield from (f"  {metric_spec}" for metric_spec in self.metrics._help())

        # Parameters and Hyperparameters
        visible_parameters = [
            parameter
            for parameter in self.parameters.values()
            if not parameter.is_hidden
        ]
        visible_hyperparameters = [
            hyperparameter
            for hyperparameter in self.hyperparameters.values()
            if not hyperparameter.is_hidden
        ]
        yield "  parameters:"
        if visible_parameters:
            for parameter in visible_parameters:
                yield from (
                    f"    {parameter_spec}" for parameter_spec in parameter._help()
                )
        else:
            yield "    No parameters"

        yield "  hyperparameters:"
        if visible_hyperparameters:
            for hyperparameter in visible_hyperparameters:
                yield from (
                    f"    {hyperparameter_spec}"
                    for hyperparameter_spec in hyperparameter._help()
                )
        else:
            yield "    No hyperparameters"


class ModelSpec(BaseModel, Helpful):
    name: str
    is_deprecated: bool
    is_hidden: bool
    keywords: list[str]
    sub_model_types: dict[str, SubModelSpec]

    model_config = ConfigDict(frozen=True)

    @property
    def model_type(self) -> str:
        return self.name

    @classmethod
    def from_dict(cls, data: datax.ModelSpecDict) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls(
            name=data["name"],
            is_deprecated=data["is_deprecated"],
            is_hidden=data["is_hidden"],
            keywords=data["keywords"],
            sub_model_types={
                (
                    sub_model_spec["name"].lower().replace("_", "").replace(" ", "")
                ): SubModelSpec.from_dict(model_type=data["name"], data=sub_model_spec)
                for sub_model_spec in data["sub_model_types"]
            },
        )

        return self  # noqa: RET504 -- seperating construction and return for clarity

    def __contains__(self, sub_model_type: str) -> bool:
        key = sub_model_type.lower().replace("_", "").replace(" ", "")
        return key in self.sub_model_types

    def __getitem__(self, sub_model_type: str) -> SubModelSpec:
        if sub_model_type not in self:
            error_msg = f"{sub_model_type.title()} sub-model does not exist"
            raise AttributeError(error_msg)
        key = sub_model_type.lower().replace("_", "").replace(" ", "")
        return self.sub_model_types[key]

    def __getattr__(self, sub_model_type: str) -> SubModelSpec:
        return self[sub_model_type]

    @property
    def types(self) -> list[str]:
        return [
            sub_model_spec.name
            for sub_model_spec in self.sub_model_types.values()
            if not sub_model_spec.is_hidden
        ]

    @override
    def _help(self) -> Generator[str]:
        # Model name
        yield f"{self.name.title()}:"
        if self.keywords:
            yield f"  keywords: {self.keywords}"

        # Sub-model types
        visible_sub_model_types = [
            sub_model_spec
            for sub_model_spec in self.sub_model_types.values()
            if not sub_model_spec.is_hidden
        ]
        if not visible_sub_model_types:
            return

        yield "  sub-model types:"
        for sub_model_spec in visible_sub_model_types:
            yield from (
                f"    {sub_model_spec}" for sub_model_spec in sub_model_spec._help()
            )


class ModelTypes(RootModel, Helpful):
    root: dict[LowercaseStr, ModelSpec]

    @classmethod
    def from_list(cls, data: list[datax.ModelSpecDict]) -> Self:
        data_dict = {model_spec["name"]: model_spec for model_spec in data}
        return cls.from_dict(data_dict)

    @classmethod
    def from_dict(cls, data: Mapping[str, Any]) -> Self:
        return cls(
            {
                model_name.lower()
                .replace("_", "")
                .replace(" ", ""): ModelSpec.from_dict(model_spec)
                for model_name, model_spec in data.items()
            }
        )

    def __len__(self) -> int:
        return len(self.root)

    def __contains__(self, name: str) -> bool:
        return name.lower().replace("_", "").replace(" ", "") in self.root

    def __getitem__(self, name: str) -> ModelSpec:
        if name not in self:
            error_msg = f"{name.title()} model does not exist"
            raise AttributeError(error_msg)
        key = name.lower().replace("_", "").replace(" ", "")
        return self.root[key]

    def __getattr__(self, name: str) -> ModelSpec:
        return self[name]

    @property
    def types(self) -> list[str]:
        return [model.name for model in self.root.values() if not model.is_hidden]

    @override
    def _help(self) -> Generator[str]:
        for model_spec in self.root.values():
            if model_spec.is_hidden:
                continue
            yield from model_spec._help()



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.specs.facet import (
    CustomFacetArgumentSpec,
    CustomFacetType,
    FacetType,
    FacetTypes,
)
from ikigai.specs.model import ModelTypes, SubModelSpec

__all__ = [
    "CustomFacetArgumentSpec",
    "CustomFacetType",
    "FacetType",
    "FacetTypes",
    "ModelTypes",
    "SubModelSpec",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.components.app import (
    App,
    AppBrowser,
    AppBuilder,
    AppDirectory,
    AppDirectoryBuilder,
)
from ikigai.components.custom_facet import (
    CustomFacet,
    CustomFacetBrowser,
    CustomFacetBuilder,
)
from ikigai.components.dataset import (
    Dataset,
    DatasetBrowser,
    DatasetBuilder,
    DatasetDirectory,
    DatasetDirectoryBuilder,
)
from ikigai.components.flow import (
    Flow,
    FlowBrowser,
    FlowBuilder,
    FlowDirectory,
    FlowDirectoryBuilder,
    Schedule,
)
from ikigai.components.flow_definition import FlowDefinitionBuilder
from ikigai.components.model import (
    Model,
    ModelBrowser,
    ModelBuilder,
    ModelDirectory,
    ModelDirectoryBuilder,
)

__all__ = [
    "App",
    "AppBrowser",
    "AppBuilder",
    "AppDirectory",
    "AppDirectoryBuilder",
    "CustomFacet",
    "CustomFacetBrowser",
    "CustomFacetBuilder",
    "Dataset",
    "DatasetBrowser",
    "DatasetBuilder",
    "DatasetDirectory",
    "DatasetDirectoryBuilder",
    "Flow",
    "FlowBrowser",
    "FlowBuilder",
    "FlowDefinitionBuilder",
    "FlowDirectory",
    "FlowDirectoryBuilder",
    "Model",
    "ModelBrowser",
    "ModelBuilder",
    "ModelDirectory",
    "ModelDirectoryBuilder",
    "Schedule",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from enum import Enum


class MissingType(Enum):
    _MISSING = "Missing"


MISSING = MissingType._MISSING



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.typing.component_browser import ComponentBrowser
from ikigai.typing.helpful import Helpful
from ikigai.typing.named_mapping import Named, NamedMapping
from ikigai.typing.protocol import (
    Directory,
    DirectoryDict,
    ModelType,
    NamedDirectoryDict,
)

__all__ = [
    "ComponentBrowser",
    "Directory",
    "DirectoryDict",
    "Helpful",
    "ModelType",
    "Named",
    "NamedDirectoryDict",
    "NamedMapping",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from collections import Counter, abc
from collections.abc import Mapping
from typing import Any, Generic, Protocol, TypeVar


class Named(Protocol):
    @property
    def name(self) -> str: ...


VT = TypeVar("VT", bound=Named)


class NamedMapping(Generic[VT], Mapping[str, VT]):
    def __init__(self, mapping: Mapping[str, VT]) -> None:
        self._names = Counter(item.name for item in mapping.values())
        self._mapping: Mapping[str, VT] = mapping

    def __getitem__(self, key: str) -> VT:
        count = self._names[key]
        if count == 0:
            raise KeyError(key)

        matches = [item for item in self._mapping.values() if item.name == key]
        if count > 1:
            error_msg = (
                f'Multiple({count}) items with name: "{key}", '
                f'use get_id(id="...") to disambiguiate between them'
            )
            raise KeyError(error_msg, matches)

        return matches[0]

    def __contains__(self, key: Any) -> bool:
        if name := getattr(key, "name", None):
            key = name
        return key in self._names

    def __iter__(self) -> abc.Iterator[str]:
        return iter(self._names)

    def __len__(self) -> int:
        return len(self._mapping)

    def __repr__(self) -> str:
        return repr(self._mapping)

    def get_id(self, id: str) -> VT:
        return self._mapping[id]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from functools import cached_property
from typing import Any

from pydantic import BaseModel, EmailStr, Field, PrivateAttr

from ikigai.client import Client
from ikigai.components.dataset import (
    Dataset,
    DatasetBrowser,
    DatasetBuilder,
    DatasetDirectory,
    DatasetDirectoryBuilder,
)
from ikigai.components.flow import (
    Flow,
    FlowBrowser,
    FlowBuilder,
    FlowDirectory,
    FlowDirectoryBuilder,
)
from ikigai.components.model import (
    Model,
    ModelBrowser,
    ModelBuilder,
    ModelDirectory,
    ModelDirectoryBuilder,
)
from ikigai.typing import (
    ComponentBrowser,
    Directory,
    NamedDirectoryDict,
    NamedMapping,
)
from ikigai.utils import AppAccessLevel, DirectoryType
from ikigai.utils.compatibility import Self, deprecated, override


class AppBuilder:
    _name: str
    _description: str
    _directory: Directory | None
    _icon: str
    _images: list[str]
    __client: Client

    def __init__(self, client: Client) -> None:
        self.__client = client
        self._name = ""
        self._description = ""
        self._directory = None
        self._icon = ""
        self._images = []

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def description(self, description: str) -> Self:
        self._description = description
        return self

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def build(self) -> App:
        app_id = self.__client.component.create_app(
            name=self._name,
            description=self._description,
            directory=self._directory,
        )
        app_dict = self.__client.component.get_app(app_id=app_id)
        return App.from_dict(data=app_dict, client=self.__client)


class AppAccess(BaseModel):
    """
    Access manager for the App.

    Grant/Update/Revoke access to the App.
    """

    __app_id: str = PrivateAttr()
    __client: Client = PrivateAttr()

    def __init__(self, *, app_id: str, client: Client) -> None:
        super().__init__()
        self.__app_id = app_id
        self.__client = client

    def grant(self, email: EmailStr, access_level: AppAccessLevel) -> Self:
        """
        Grant access to the App.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to grant access to.

        access_level: AppAccessLevel
            Access level to grant to the user.

        Returns
        -------

        Self
            Access manager for the App.
        """
        self.__client.component.grant_app_access(
            app_id=self.__app_id,
            email=email,
            access_level=access_level,
        )
        return self

    def update(self, email: EmailStr, access_level: AppAccessLevel) -> Self:
        """
        Update access to the App.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to update access for.

        access_level: AppAccessLevel
            New access level for the user.

        Returns
        -------

        Self
            Access manager for the App.
        """
        self.__client.component.update_app_access(
            app_id=self.__app_id,
            email=email,
            access_level=access_level,
        )
        return self

    def revoke(self, email: EmailStr) -> Self:
        """
        Revoke access to the App.

        Parameters
        ----------

        email: EmailStr
            Email address of the user to revoke access for.

        Returns
        -------

        Self
            Access manager for the App.
        """
        self.__client.component.revoke_app_access(
            app_id=self.__app_id,
            email=email,
        )
        return self


class App(BaseModel):
    """
    Represents an App in the Ikigai platform.

    Apps are the basic organizational units in Ikigai.

    Attributes
    ----------

    app_id: str
        Unique identifier of the app.

    name: str
        Name of the App.

    owner: EmailStr
        Email address of the app owner.

    description: str
        Description of the app.

    created_at: datetime
        Timestamp when the app was created.

    modified_at: datetime
        Timestamp when the app was last modified.

    last_used_at: datetime
        Timestamp when the app was last used.
    """

    app_id: str = Field(validation_alias="project_id")
    name: str
    owner: EmailStr
    description: str
    created_at: datetime
    modified_at: datetime
    last_used_at: datetime
    __client: Client = PrivateAttr()

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        self = cls.model_validate(data)
        self.__client = client
        return self

    """
    Operations on App
    """

    def to_dict(self) -> dict:
        """
        Convert the App to a dictionary representation.

        Returns
        -------

        dict
            Dictionary containing the App attributes.
        """
        return {
            "app_id": self.app_id,
            "name": self.name,
            "owner": self.owner,
            "description": self.description,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
            "last_used_at": self.last_used_at,
        }

    def delete(self) -> None:
        """
        Delete the App.

        Returns
        -------

        None
        """
        self.__client.component.delete_app(app_id=self.app_id)
        return None

    def rename(self, name: str) -> Self:
        """
        Rename the App.

        Parameters
        ----------

        name: str
            New name for the App.

        Returns
        -------

        Self
            Updated App.
        """
        _ = self.__client.component.edit_app(app_id=self.app_id, name=name)
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        """
        Move the App to the target directory.

        Parameters
        ----------

        directory: Directory
            Target directory.

        Returns
        -------

        Self
            Updated App.
        """
        _ = self.__client.component.edit_app(app_id=self.app_id, directory=directory)
        return self

    def update_description(self, description: str) -> Self:
        """
        Update the App's description.

        Parameters
        ----------

        description: str
            The new description for the App.

        Returns
        -------

        Self
            The updated App with the new description.
        """
        _ = self.__client.component.edit_app(
            app_id=self.app_id, description=description
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.description = description
        return self

    @cached_property
    def access(self) -> AppAccess:
        """
        Manage access to the App.

        Returns
        -------

        AppAccess
            Access manager for the App.
        """
        return AppAccess(app_id=self.app_id, client=self.__client)

    def describe(self) -> dict[str, Any]:
        """
        Get details about the App including its components.

        Returns
        -------

        dict[str, Any]
            Dictionary containing App details and its associated
            components.
        """
        components = self.__client.component.get_components_for_app(app_id=self.app_id)

        # Combine components information with app info
        return {
            "app": self.to_dict(),
            "components": components,
        }

    """
    Access Components in the App
    """

    @property
    def datasets(self) -> ComponentBrowser[Dataset]:
        """
        Access the datasets associated with the App.

        Returns a `ComponentBrowser[Dataset]` to access this app's datasets.

        Returns
        -------

        DatasetBrowser
            Dataset browser for interacting with the App's datasets.

        Examples
        --------

        Use `search(query: str)`, which returns datasets matching a query
        string.

        >>> datasets = app.datasets
        >>> dataset = datasets.search("Examp")

        Retrieve a dataset by name.

        >>> datasets = app.datasets
        >>> dataset = datasets['Example Dataset']
        """
        return DatasetBrowser(app_id=self.app_id, client=self.__client)

    @property
    def dataset(self) -> DatasetBuilder:
        return DatasetBuilder(client=self.__client, app_id=self.app_id)

    def dataset_directories(self) -> NamedMapping[DatasetDirectory]:
        directory_dicts = self.__client.component.get_dataset_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                DatasetDirectory.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def dataset_directory(self) -> DatasetDirectoryBuilder:
        return DatasetDirectoryBuilder(client=self.__client, app_id=self.app_id)

    @property
    def flows(self) -> ComponentBrowser[Flow]:
        """
        Access the flows associated with the App.

        Returns a `ComponentBrowser[Flow]` to access this app's flows.

        Returns
        -------

        FlowBrowser
            Flow browser for interacting with the App's flows.

        Examples
        --------

        Use `search(query: str)` to retrieve flows matching a query.

        >>> flows = app.flows
        >>> flow = flows.search("Examp")

        Retrieve a flow by name.

        >>> flows = app.flows
        >>> flow = flows['Example Flow']
        """
        return FlowBrowser(app_id=self.app_id, client=self.__client)

    @property
    def flow(self) -> FlowBuilder:
        return FlowBuilder(client=self.__client, app_id=self.app_id)

    def flow_directories(self) -> NamedMapping[FlowDirectory]:
        directory_dicts = self.__client.component.get_flow_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                FlowDirectory.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def flow_directory(self) -> FlowDirectoryBuilder:
        return FlowDirectoryBuilder(client=self.__client, app_id=self.app_id)

    @property
    def models(self) -> ComponentBrowser[Model]:
        """
        Access the models associated with the App.

        Returns a `ComponentBrowser[Model]` to access this app's models.

        Returns
        -------

        ModelBrowser
            Model browser for interacting with the App's models.

        Examples
        --------

        Use `search(query: str)` to retrieve flows matching a query.

        >>> models = app.models
        >>> model = models.search("model")

        Retrieve a model by name.

        >>> models = app.models
        >>> model = models['model_name']
        """
        return ModelBrowser(app_id=self.app_id, client=self.__client)

    @property
    def model(self) -> ModelBuilder:
        return ModelBuilder(client=self.__client, app_id=self.app_id)

    def model_directories(self) -> NamedMapping[ModelDirectory]:
        directory_dicts = self.__client.component.get_model_directories_for_app(
            app_id=self.app_id
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                ModelDirectory.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def model_directory(self) -> ModelDirectoryBuilder:
        return ModelDirectoryBuilder(client=self.__client, app_id=self.app_id)


class AppBrowser(ComponentBrowser[App]):
    __client: Client

    def __init__(self, client: Client) -> None:
        self.__client = client

    @deprecated("Prefer directly loading by name:\n\tikigai.apps['app_name']")
    @override
    def __call__(self) -> NamedMapping[App]:
        apps = {
            app["project_id"]: App.from_dict(data=app, client=self.__client)
            for app in self.__client.component.get_apps_for_user()
        }

        return NamedMapping(apps)

    @override
    def __getitem__(self, name: str) -> App:
        app_dict = self.__client.component.get_app_by_name(name)
        return App.from_dict(data=app_dict, client=self.__client)

    @override
    def search(self, query: str) -> NamedMapping[App]:
        matching_apps = {
            app["project_id"]: App.from_dict(data=app, client=self.__client)
            for app in self.__client.search.search_projects_for_user(query=query)
        }

        return NamedMapping(matching_apps)


class AppDirectoryBuilder:
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client) -> None:
        self.__client = client
        self._name = ""
        self._parent = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def parent(self, parent: Directory) -> Self:
        self._parent = parent
        return self

    def build(self) -> AppDirectory:
        directory_id = self.__client.component.create_app_directory(
            name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_app_directory(
            directory_id=directory_id
        )
        return AppDirectory.from_dict(data=directory_dict, client=self.__client)


class AppDirectory(BaseModel):
    directory_id: str
    name: str
    __client: Client = PrivateAttr()

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.APP

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def delete(self) -> None:
        self.__client.component.delete_app_directory(directory_id=self.directory_id)
        return None

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_app_directories_for_user(
            directory_id=self.directory_id,
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def apps(self) -> NamedMapping[App]:
        app_dicts = self.__client.component.get_apps_for_user(
            directory_id=self.directory_id
        )

        apps = {
            app.app_id: app
            for app in (
                App.from_dict(data=app_dict, client=self.__client)
                for app_dict in app_dicts
            )
        }

        return NamedMapping(apps)



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.utils.enums import (
    AppAccessLevel,
    CustomFacetAccessLevel,
    CustomFacetArgumentType,
    DatasetDataType,
    DatasetDownloadStatus,
    DirectoryType,
    FacetArgumentType,
    FlowStatus,
    ModelHyperparameterType,
    ModelParameterType,
)

__all__: list[str] = [
    "AppAccessLevel",
    "CustomFacetAccessLevel",
    "CustomFacetArgumentType",
    "DatasetDataType",
    "DatasetDownloadStatus",
    "DirectoryType",
    "FacetArgumentType",
    "FlowStatus",
    "ModelHyperparameterType",
    "ModelParameterType",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any

__ONE_HOUR_AS_MINUTES = 60
__CRON_PARTS = ["minute", "hour", "day", "month", "day_of_week"]


def optional_str(value: Any) -> str | None:
    if not value:
        return None
    return str(value)


def cron_str(value: Any) -> str:
    if not isinstance(value, str):
        error_msg = "Cron must be a string"
        raise TypeError(error_msg)

    cron_components = value.strip().split(" ")

    # Validate the number of components
    if len(cron_components) != len(__CRON_PARTS):
        error_msg = (
            "Incorrect cron format, expected format: minute hour day month day_of_week"
            f"but got {value}"
        )
        raise ValueError(error_msg)

    # On Ikigai, the minute field must be an integer between 0 and 59
    minute = cron_components[0]
    if not minute.isdigit():
        error_msg = "Minute field must be an integer"
        raise TypeError(error_msg)

    if not (0 <= int(minute) < __ONE_HOUR_AS_MINUTES):
        error_msg = "Minute field must be an integer between 0 and 59"
        raise ValueError(error_msg)

    return value.strip()



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from ikigai import Ikigai


def test_facet_types_property_access(ikigai: Ikigai) -> None:
    facet_types = ikigai.facet_types
    assert facet_types

    assert hasattr(facet_types, "INPUT")
    assert hasattr(facet_types, "MID")
    assert hasattr(facet_types, "OUTPUT")


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


def test_iplt_10504_multi_word_facet_type_access(ikigai: Ikigai) -> None:
    facet_types = ikigai.facet_types

    # Check that various casings and formats for Custom Facet work
    assert "Custom Facet" in facet_types.INPUT
    assert "CUSTOM FACET" in facet_types.INPUT
    assert "custom_facet" in facet_types.INPUT
    assert "CUSTOM_FACET" in facet_types.INPUT

    assert facet_types.INPUT.CUSTOM_FACET
    assert facet_types.INPUT.custom_facet
    assert facet_types.INPUT.CustomFacet
    assert facet_types.INPUT["Custom Facet"]

    # Check that various casings and formats for Drop Columns work
    assert "Drop Columns" in facet_types.MID
    assert "DROP COLUMNS" in facet_types.MID
    assert "drop_columns" in facet_types.MID
    assert "DROP_COLUMNS" in facet_types.MID

    assert facet_types.MID.DROP_COLUMNS
    assert facet_types.MID.drop_columns
    assert facet_types.MID.DropColumns
    assert facet_types.MID["Drop Columns"]



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.typing.pydantic_extensions.types import (
    CronStr,
    LowercaseStr,
    OptionalStr,
    TimestampSerializableDatetime,
    TimestampSerializableOptionalDatetime,
)

__all__: list[str] = [
    "CronStr",
    "LowercaseStr",
    "OptionalStr",
    "TimestampSerializableDatetime",
    "TimestampSerializableOptionalDatetime",
]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import io
import logging
import math
import time
from collections.abc import Mapping
from datetime import datetime
from http import HTTPStatus
from pathlib import Path
from typing import Any

import pandas as pd
import requests
from pydantic import BaseModel, Field, PrivateAttr

from ikigai.client import Client, datax
from ikigai.typing import ComponentBrowser, Directory, NamedDirectoryDict, NamedMapping
from ikigai.utils import DatasetDataType, DatasetDownloadStatus, DirectoryType
from ikigai.utils.compatibility import Self, deprecated, override

logger = logging.getLogger("ikigai.components")


def __upload_data(
    client: Client,
    app_id: str,
    dataset_id: str,
    data: bytes,
    filename: str,
) -> None:
    file_size = len(data)
    multipart_upload_metadata = client.component.get_dataset_multipart_upload_urls(
        dataset_id=dataset_id,
        app_id=app_id,
        filename=filename,
        file_size=file_size,
    )

    content_type = multipart_upload_metadata["content_type"]
    upload_urls = multipart_upload_metadata["urls"]
    upload_id = multipart_upload_metadata["upload_id"]

    num_chunks = len(upload_urls)
    chunk_size = math.ceil(file_size / num_chunks)

    etags: dict[int, str] = {}
    try:
        with requests.session() as request:
            request.headers.update(
                {"Content-Type": content_type, "Cache-Control": "no-cache"}
            )
            for idx, (chunk_idx, upload_url) in enumerate(sorted(upload_urls.items())):
                chunk_start, chunk_end = (idx * chunk_size, (idx + 1) * chunk_size)
                chunk = data[chunk_start : min(chunk_end, file_size)]
                resp = request.put(url=upload_url, data=chunk)
                if resp.status_code != HTTPStatus.OK:
                    error_msg = (
                        f"Failed to upload chunk {chunk_idx:02d} of {num_chunks:02d} "
                        "received response:\n"
                        f"[{resp.status_code}] {resp.text}"
                    )
                    raise RuntimeError(error_msg)

                # Get etags from response header
                etags[chunk_idx] = resp.headers["ETag"]
    except Exception:
        client.component.abort_datset_multipart_upload(
            app_id=app_id,
            dataset_id=dataset_id,
            filename=filename,
            upload_id=upload_id,
        )
        raise

    # Complete Dataset upload
    client.component.complete_datset_multipart_upload(
        app_id=app_id,
        dataset_id=dataset_id,
        filename=filename,
        upload_id=upload_id,
        etags=etags,
    )


def _upload_data(
    client: Client, app_id: str, dataset_id: str, name: str, data: bytes
) -> None:
    if not data:
        error_msg = "Dataset is empty"
        raise ValueError(error_msg)

    filename = f"{name}.csv"

    __upload_data(
        client=client,
        app_id=app_id,
        dataset_id=dataset_id,
        data=data,
        filename=filename,
    )

    upload_completion_time = time.time()
    client.component.verify_dataset_upload(
        app_id=app_id, dataset_id=dataset_id, filename=filename
    )

    dataset_status: str = "RUNNING"
    while dataset_status == "RUNNING":
        # Block thread while dataset is still being processed
        time.sleep(1)
        dataset_status = client.component.confirm_dataset_upload(
            app_id=app_id, dataset_id=dataset_id
        )

    if dataset_status != "SUCCESS":
        error_msg = f"Dataset upload failed, upload ended with status {dataset_status}"
        raise RuntimeError(error_msg)

    dataset_logs = client.component.get_dataset_log(
        app_id=app_id, dataset_id=dataset_id, limit=5
    )
    dataset_verified = [
        log
        for log in dataset_logs
        if log["status"] == "SUCCESS"
        and int(log["timestamp"]) > upload_completion_time
        and log["job_type"] == "UPLOAD_DATASET"
    ]
    if not dataset_verified:
        # TODO: Improve Error
        error_msg = "Dataset failed to verify"
        raise RuntimeError(error_msg, dataset_logs)


def _get_dataset_download_url(client: Client, app_id: str, dataset_id: str) -> str:
    response = client.component.initialize_dataset_download(
        app_id=app_id,
        dataset_id=dataset_id,
    )

    while response["status"] == DatasetDownloadStatus.IN_PROGRESS:
        time.sleep(1)
        response = client.component.initialize_dataset_download(
            app_id=app_id,
            dataset_id=dataset_id,
        )

    if response["status"] == DatasetDownloadStatus.FAILED:
        error_msg = (
            "Dataset download failed, dataset size might be too large to download"
        )
        raise RuntimeError(error_msg)

    if response["status"] != DatasetDownloadStatus.SUCCESS:
        error_msg = f"Got unexpected dataset download status: {response['status']}"
        raise RuntimeError(error_msg)

    return response["url"]


class DatasetBuilder:
    _app_id: str
    _name: str
    _data: bytes | None
    _directory: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._data = None
        self._directory = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def df(self, data: pd.DataFrame) -> Self:
        buffer = io.BytesIO()
        data.to_csv(buffer, index_label=False, index=False)
        self._data = buffer.getvalue()
        return self

    def csv(self, file: Path | str) -> Self:
        if isinstance(file, str):
            file = Path(file)
        with file.open("rb") as fp:
            self._data = fp.read()
        return self

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def build(self) -> Dataset:
        if self._data is None:
            error_msg = "Dataset is empty"
            raise ValueError(error_msg)

        dataset_id = self.__client.component.create_dataset(
            app_id=self._app_id, name=self._name, directory=self._directory
        )

        try:
            _upload_data(
                client=self.__client,
                app_id=self._app_id,
                dataset_id=dataset_id,
                name=self._name,
                data=self._data,
            )
        except Exception:
            # Delete created record and re-raise
            self.__client.component.delete_dataset(
                app_id=self._app_id, dataset_id=dataset_id
            )
            raise

        # Populate Dataset object
        dataset_dict = self.__client.component.get_dataset(
            app_id=self._app_id, dataset_id=dataset_id
        )

        return Dataset.from_dict(data=dataset_dict, client=self.__client)


class ColumnDataType(BaseModel):
    data_type: DatasetDataType
    data_formats: dict[str, str]


class Dataset(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    dataset_id: str
    name: str
    filename: str
    file_extension: str
    data_types: dict[str, ColumnDataType]
    size: int
    created_at: datetime
    modified_at: datetime
    __client: Client = PrivateAttr()

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> dict:
        return {
            "dataset_id": self.dataset_id,
            "name": self.name,
            "filename": self.filename,
            "file_extension": self.file_extension,
            "data_types": self.data_types,
            "size": self.size,
            "created_at": self.created_at,
            "modified_at": self.modified_at,
        }

    def delete(self) -> None:
        self.__client.component.delete_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id
        )
        return None

    def rename(self, name: str) -> Self:
        _ = self.__client.component.edit_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id, name=name
        )
        # TODO: handle error case, currently it is a raise NotImplemented from Session
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        _ = self.__client.component.edit_dataset(
            app_id=self.app_id,
            dataset_id=self.dataset_id,
            directory=directory,
        )
        return self

    def df(self, **parser_options) -> pd.DataFrame:
        download_url = _get_dataset_download_url(
            client=self.__client,
            app_id=self.app_id,
            dataset_id=self.dataset_id,
        )
        return pd.read_csv(download_url, **parser_options)

    def edit_data(self, data: pd.DataFrame) -> None:
        buffer = io.BytesIO()
        data.to_csv(buffer, index_label=False, index=False)

        _upload_data(
            client=self.__client,
            app_id=self.app_id,
            dataset_id=self.dataset_id,
            name=self.name,
            data=buffer.getvalue(),
        )

    def describe(self) -> datax.DatasetDict:
        return self.__client.component.get_dataset(
            app_id=self.app_id, dataset_id=self.dataset_id
        )


class DatasetBrowser(ComponentBrowser[Dataset]):
    __app_id: str
    __client: Client

    def __init__(self, app_id: str, client: Client) -> None:
        self.__app_id = app_id
        self.__client = client

    @deprecated("Prefer directly loading by name:\n\tapp.datasets['dataset_name']")
    @override
    def __call__(self) -> NamedMapping[Dataset]:
        datasets = {
            dataset["dataset_id"]: Dataset.from_dict(data=dataset, client=self.__client)
            for dataset in self.__client.component.get_datasets_for_app(
                app_id=self.__app_id
            )
        }

        return NamedMapping(datasets)

    @override
    def __getitem__(self, name: str) -> Dataset:
        dataset_dict = self.__client.component.get_dataset_by_name(
            app_id=self.__app_id, name=name
        )

        return Dataset.from_dict(data=dataset_dict, client=self.__client)

    @override
    def search(self, query: str) -> NamedMapping[Dataset]:
        matching_datasets = {
            dataset["dataset_id"]: Dataset.from_dict(data=dataset, client=self.__client)
            for dataset in self.__client.search.search_datasets_for_project(
                app_id=self.__app_id, query=query
            )
        }

        return NamedMapping(matching_datasets)


class DatasetDirectoryBuilder:
    _app_id: str
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._parent = None

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def parent(self, parent: Directory) -> Self:
        self._parent = parent
        return self

    def build(self) -> DatasetDirectory:
        directory_id = self.__client.component.create_dataset_directory(
            app_id=self._app_id, name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_dataset_directory(
            app_id=self._app_id, directory_id=directory_id
        )

        return DatasetDirectory.from_dict(data=directory_dict, client=self.__client)


class DatasetDirectory(BaseModel):
    app_id: str = Field(validation_alias="project_id")
    directory_id: str
    name: str
    __client: Client = PrivateAttr()

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.DATASET

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_dataset_directories_for_app(
            app_id=self.app_id, parent=self
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def datasets(self) -> NamedMapping[Dataset]:
        dataset_dicts = self.__client.component.get_datasets_for_app(
            app_id=self.app_id,
            directory_id=self.directory_id,
        )
        datasets = {
            dataset.dataset_id: dataset
            for dataset in (
                Dataset.from_dict(data=dataset_dict, client=self.__client)
                for dataset_dict in dataset_dicts
            )
        }

        return NamedMapping(datasets)



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import InitVar

from pydantic import EmailStr, Field, HttpUrl
from pydantic.dataclasses import dataclass

from ikigai import components, specs
from ikigai.client import Client, SSLConfig
from ikigai.typing import ComponentBrowser, NamedMapping
from ikigai.utils.compatibility import deprecated
from ikigai.utils.missing import MISSING, MissingType


@dataclass
class Ikigai:
    """
    Main Ikigai class to interact with the Ikigai platform.

    Parameters
    ----------

    user_email: str
        Email of the user.

    api_key: str
        API key for authentication.

    base_url: str
        Base URL of the Ikigai API endpoints. Default is
        "https://api.ikigailabs.io".

    ssl: bool or str or tuple
        SSL configuration. See `ikigai.utils.config` for more details.
        Set to `False` to disable SSL verification (unsafe), or provide
        custom SSL certificate by providing the path to a certificate
        (.pem) file or a tuple of (certificate, key).
    """

    user_email: EmailStr
    api_key: InitVar[str]
    base_url: HttpUrl = Field(default=HttpUrl("https://api.ikigailabs.io"))
    ssl: InitVar[SSLConfig | MissingType] = MISSING
    __client: Client = Field(init=False)

    def __post_init__(
        self, api_key: str, ssl: SSLConfig | MissingType = MISSING
    ) -> None:
        if ssl is MISSING:
            ssl = True
        self.__client = Client(
            user_email=self.user_email, api_key=api_key, base_url=self.base_url, ssl=ssl
        )

    @property
    def apps(self) -> ComponentBrowser[components.App]:
        """
        Access the Apps associated with the user.

        Returns a `ComponentBrowser[App]` to access the user's apps.

        Returns
        -------

        ComponentBrowser[components.App]
            Browser for Apps.

        Examples
        --------

        Use `search(query: str)`, which returns Apps matching a query
        string.

        >>> apps = ikigai.apps()
        >>> app = apps.search("Examp")

        Retrieve an App by name.

        >>> apps = ikigai.apps()
        >>> app = apps["Example App"]
        """
        return components.AppBrowser(client=self.__client)

    @property
    def app(self) -> components.AppBuilder:
        """
        Get a Builder to create a new App.

        Returns
        -------

        components.AppBuilder
            A new App builder.
        """
        return components.AppBuilder(client=self.__client)

    @property
    def custom_facets(self) -> ComponentBrowser[components.CustomFacet]:
        """
        Access the custom facets associated with the user.

        Returns a `ComponentBrowser[CustomFacet]` to access the user's custom facets.

        Returns
        -------

        ComponentBrowser[components.CustomFacet]
            Browser for Custom Facets.

        Examples
        --------

        Use `search(query: str)` to retrieve custom facets matching a query.

        >>> custom_facets = ikigai.custom_facets
        >>> custom_facet = custom_facets.search("Examp")

        Retrieve a custom facet by name.

        >>> custom_facets = ikigai.custom_facets
        >>> custom_facet = custom_facets["Example Custom Facet"]
        """
        return components.CustomFacetBrowser(client=self.__client)

    @property
    def custom_facet(self) -> components.CustomFacetBuilder:
        """
        Get a Builder to create a new Custom Facet.

        Returns
        -------

        components.CustomFacetBuilder
            A new Custom Facet builder.
        """
        return components.CustomFacetBuilder(client=self.__client)

    @deprecated("Use ikigai.app_directories() instead, will be removed in v0.4.0")
    def directories(self) -> NamedMapping[components.AppDirectory]:
        return self.app_directories()

    def app_directories(self) -> NamedMapping[components.AppDirectory]:
        """
        Get all App Directories for the user.

        Returns
        -------

        NamedMapping[components.AppDirectory]
            Mapping of names to Directories.
        """
        directory_dicts = self.__client.component.get_app_directories_for_user()
        directories = {
            directory.directory_id: directory
            for directory in (
                components.AppDirectory.from_dict(
                    data=directory_dict, client=self.__client
                )
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    @property
    def app_directory(self) -> components.AppDirectoryBuilder:
        """
        Builder to create a new App Directory

        Returns
        -------
        components.AppDirectoryBuilder
            A new App Directory builder object
        """
        return components.AppDirectoryBuilder(client=self.__client)

    @property
    def builder(self) -> components.FlowDefinitionBuilder:
        """
        Get a Builder to create a new Flow Definition.

        Returns
        -------

        components.FlowDefinitionBuilder
            A new Flow Definition builder.
        """
        return components.FlowDefinitionBuilder()

    @property
    def facet_types(self) -> specs.FacetTypes:
        """
        Available facets for use on the Ikigai platform.

        Returns
        -------

        components.FacetTypes
            Available facet types.
        """
        return specs.FacetTypes.from_dict(
            data=self.__client.component.get_facet_specs()
        )

    @property
    def model_types(self) -> specs.ModelTypes:
        """
        Available model types in the Ikigai platform.

        Returns
        -------

        components.ModelTypes
            Available model types.
        """
        return specs.ModelTypes.from_list(
            data=self.__client.component.get_model_specs()
        )



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import abc
from typing import Generic, TypeVar

from ikigai.typing.named_mapping import Named, NamedMapping

T = TypeVar("T", bound=Named)


class ComponentBrowser(Generic[T], abc.ABC):
    @abc.abstractmethod
    def __call__(self) -> NamedMapping[T]:
        """
        Get as many components as possible

        Note
        ----
            Usage of this method is discouraged as the platform may truncate
            the list of components returned if it exceeds a certain limit.

        Returns
        -------
        NamedMapping[T]
            A mapping of component names to components
        """

    @abc.abstractmethod
    def __getitem__(self, name: str) -> T:
        """
        Get a component by name

        Parameters
        ----------
        name : str
            Name of the component to get

        Returns
        -------
        T
            The component with the given name

        Raises
        ------
        KeyError
            If the component with the given name does not exist
        """

    @abc.abstractmethod
    def search(self, query: str) -> NamedMapping[T]:
        """
        Search for a component by name

        Parameters
        ----------
        query : str
            Query string to search for

        Returns
        -------
        NamedMapping[T]
            A mapping of component names to components that match the query
        """



# SPDX-FileCopyrightText: 2025-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from contextlib import ExitStack

import pytest

from ikigai import Ikigai


def test_model_types(
    ikigai: Ikigai,
    app_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = ikigai.model_types
    assert model_types is not None
    assert len(model_types) > 0
    lasso = model_types["Linear"]["Lasso"]

    assert lasso is not None
    assert lasso.model_type == "Linear"
    assert lasso.sub_model_type == "Lasso"


def test_model_creation(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    models = app.models()
    assert len(models) == 0

    model_types = ikigai.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    models_after_creation = app.models()
    assert len(models_after_creation) == 1
    assert models_after_creation[model.name]

    model.delete()
    models_after_deletion = app.models()
    assert len(models_after_deletion) == 0

    with pytest.raises(KeyError):
        models_after_deletion[model.name]
    assert model.model_id not in models_after_deletion


def test_model_editing(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = ikigai.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    model.rename(f"updated {model_name}")
    model.update_description("updated description")

    model_after_edit = app.models().get_id(model.model_id)
    assert model_after_edit.name == model.name
    assert model_after_edit.name == f"updated {model_name}"
    assert model_after_edit.description == model.description
    assert model_after_edit.description == "updated description"
    assert model_after_edit.model_type == model.model_type
    assert model_after_edit.model_type == "Linear"
    assert model_after_edit.sub_model_type == model.sub_model_type
    assert model_after_edit.sub_model_type == "Lasso"


def test_model_describe(
    ikigai: Ikigai,
    app_name: str,
    model_name: str,
    cleanup: ExitStack,
) -> None:
    app = ikigai.app.new(name=app_name).description("A test app").build()
    cleanup.callback(app.delete)

    model_types = ikigai.model_types
    model = (
        app.model.new(model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    model_description = model.describe()
    assert model_description is not None
    assert model_description["name"] == model.name
    assert model_description["description"] == model.description


def test_model_browser_1(
    ikigai: Ikigai, app_name: str, model_name: str, cleanup: ExitStack
) -> None:
    app = ikigai.app.new(name=app_name).description("Test to get model by name").build()
    cleanup.callback(app.delete)

    model_types = ikigai.model_types
    model = (
        app.model.new(name=model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    fetched_model = app.models[model_name]
    assert fetched_model.model_id == model.model_id
    assert fetched_model.name == model_name


def test_model_browser_search_1(
    ikigai: Ikigai, app_name: str, model_name: str, cleanup: ExitStack
) -> None:
    app = ikigai.app.new(name=app_name).description("Test to get model by name").build()
    cleanup.callback(app.delete)

    model_types = ikigai.model_types
    model = (
        app.model.new(name=model_name)
        .model_type(model_type=model_types["Linear"]["Lasso"])
        .build()
    )

    model_name_substr = model_name.split("-", maxsplit=1)[1]
    fetched_models = app.models.search(model_name_substr)

    assert model_name in fetched_models
    fetched_model = fetched_models[model_name]

    assert fetched_model.model_id == model.model_id


def test_model_builder_hyperparameter_groups(
    ikigai: Ikigai,
) -> None:
    flow_builder = ikigai.builder
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    expected_model_facet_args = {
        "model_name": "test_model_name_1",
        "parameters": {
            "time_column": "WEEK_END_DT",
            "identifier_columns": ["forecast_identifier"],
            "value_column": "MEASURE_QTY",
            "mode": "train",
        },
        "processing": {
            "type": "base",
            "return_all_levels": False,
            "fill_missing_values": "zero",
            "drop_threshold": 0.9,
            "include_reals": True,
            "nonnegative": True,
        },
        "model_selection": {
            "models_to_include": ["Additive", "Sma", "Tsfm0"],
            "eval_method": "holdout",
            "time_budget": 100,
            "enable_parallel_processing": True,
            "best_model_only": True,
            "confidence": 0.7,
            "enable_conformal_interval": False,
        },
        "metrics": [
            {
                "interval_to_predict": 10,
                "holdout_period": 10,
                "metric": "weighted_mean_absolute_percentage_error",
            }
        ],
    }
    aicast_facet = (
        flow_builder.model_facet(
            facet_type=facet_types.MID.AI_CAST, model_type=model_types.AI_CAST.base
        )
        .arguments(
            model_name="test_model_name_1",
        )
        .parameters(
            time_column="WEEK_END_DT",
            identifier_columns=["forecast_identifier"],
            value_column="MEASURE_QTY",
            mode="train",
        )
        .hyperparameters(
            type="base",
            models_to_include=["Additive", "Sma", "Tsfm0"],
            eval_method="holdout",
            time_budget=100,
            enable_parallel_processing=True,
            best_model_only=True,
            confidence=0.7,
        )
        .hyperparameters(
            enable_conformal_interval=False,
            return_all_levels=False,
            fill_missing_values="zero",
            drop_threshold=0.9,
            include_reals=True,
            nonnegative=True,
            interval_to_predict=10,
            holdout_period=10,
            metric="weighted_mean_absolute_percentage_error",
        )
    ).build()
    assert aicast_facet.facets[0].arguments == expected_model_facet_args


def test_model_builder_hyperparameter_typechecking_invalid_name(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="is not valid for Lasso models"):
        # bad_hyperparam is not a valid hyperparameter for Lasso
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types["Linear"]["Lasso"],
        ).hyperparameters(bad_hyperparam=123)


def test_model_builder_hyperparameter_typechecking_options(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="must be one of"):
        # embedding model type must be one of few options
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types.VECTORIZER.OPENAI,
        ).hyperparameters(embedding_model="bad_embedding_model")


def test_model_builder_hyperparameter_typechecking_number(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(TypeError, match="must be numeric"):
        # alpha should be numeric, not str
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types["Linear"]["Lasso"],
        ).hyperparameters(alpha="0.1")


def test_model_builder_hyperparameter_typechecking_boolean(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(TypeError, match="must be boolean"):
        # use_scaling must be boolean, not str
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types.EMBEDDING.LGMT1,
        ).hyperparameters(use_scaling="True")


def test_model_builder_parameter_typechecking_invalid_name(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(ValueError, match="is not valid for Lasso models"):
        # bad_param is not a valid parameter for Lasso
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types.LINEAR.LASSO,
        ).parameters(bad_param="value")


def test_model_builder_parameter_typechecking_text(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(TypeError, match="must be string"):
        # target_column should be string, not int
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types.LINEAR.LASSO,
        ).parameters(target_column=123)


def test_model_builder_parameter_typechecking_list(
    ikigai: Ikigai,
) -> None:
    facet_types = ikigai.facet_types
    model_types = ikigai.model_types
    builder = ikigai.builder
    with pytest.raises(TypeError, match="must be list"):
        # feature_columns should be a list, not a string
        builder.facet(facet_type=facet_types.INPUT.IMPORTED).model_facet(
            facet_type=facet_types.MID.PREDICT,
            model_type=model_types.AI_CAST.BASE,
        ).parameters(identifier_columns="not-a-list")



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
from collections.abc import Mapping
from datetime import datetime
from typing import Any

from pydantic import AliasChoices, BaseModel, Field, PrivateAttr

from ikigai.client import Client
from ikigai.specs import SubModelSpec
from ikigai.typing import ComponentBrowser, Directory, NamedDirectoryDict, NamedMapping
from ikigai.utils import DirectoryType
from ikigai.utils.compatibility import Self, deprecated, override

logger = logging.getLogger("ikigai.components")


class ModelBuilder:
    _app_id: str
    _name: str
    _directory: Directory | None
    _model_type: SubModelSpec | None
    _description: str
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._directory = None
        self._model_type = None
        self._description = ""

    def new(self, name: str) -> Self:
        self._name = name
        return self

    def directory(self, directory: Directory) -> Self:
        self._directory = directory
        return self

    def model_type(self, model_type: SubModelSpec) -> Self:
        self._model_type = model_type
        return self

    def description(self, description: str) -> Self:
        self._description = description
        return self

    def build(self) -> Model:
        if self._model_type is None:
            error_msg = "Model type must be specified"
            raise ValueError(error_msg)

        model_id = self.__client.component.create_model(
            app_id=self._app_id,
            name=self._name,
            directory=self._directory,
            model_type=self._model_type,
            description=self._description,
        )
        # Populate the model object
        model_dict = self.__client.component.get_model(
            app_id=self._app_id, model_id=model_id
        )

        return Model.from_dict(data=model_dict, client=self.__client)


class Model(BaseModel):
    app_id: str = Field(validation_alias=AliasChoices("app_id", "project_id"))
    model_id: str
    name: str
    model_type: str
    sub_model_type: str
    description: str
    created_at: datetime
    modified_at: datetime
    __client: Client = PrivateAttr()

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def delete(self) -> None:
        self.__client.component.delete_model(app_id=self.app_id, model_id=self.model_id)
        return None

    def rename(self, name: str) -> Self:
        self.__client.component.edit_model(
            app_id=self.app_id, model_id=self.model_id, name=name
        )
        self.name = name
        return self

    def move(self, directory: Directory) -> Self:
        self.__client.component.edit_model(
            app_id=self.app_id, model_id=self.model_id, directory=directory
        )
        return self

    def update_description(self, description: str) -> Self:
        self.__client.component.edit_model(
            app_id=self.app_id, model_id=self.model_id, description=description
        )
        self.description = description
        return self

    def versions(self) -> NamedMapping[ModelVersion]:
        version_dicts = self.__client.component.get_model_versions(
            app_id=self.app_id, model_id=self.model_id
        )
        versions = {
            version.version_id: version
            for version in (
                ModelVersion.from_dict(
                    app_id=self.app_id, data=version_dict, client=self.__client
                )
                for version_dict in version_dicts
            )
        }

        return NamedMapping(versions)

    def describe(self) -> Mapping[str, Any]:
        return self.__client.component.get_model(
            app_id=self.app_id, model_id=self.model_id
        )


class ModelBrowser(ComponentBrowser[Model]):
    __app_id: str
    __client: Client

    def __init__(self, app_id: str, client: Client) -> None:
        self.__app_id = app_id
        self.__client = client

    @deprecated("Prefer directly loading by name:\n\tapp.models['model_name']")
    @override
    def __call__(self) -> NamedMapping[Model]:
        models = {
            model["model_id"]: Model.from_dict(data=model, client=self.__client)
            for model in self.__client.component.get_models_for_app(
                app_id=self.__app_id
            )
        }

        return NamedMapping(models)

    @override
    def __getitem__(self, name: str) -> Model:
        model_dict = self.__client.component.get_model_by_name(
            app_id=self.__app_id, name=name
        )

        return Model.from_dict(data=model_dict, client=self.__client)

    @override
    def search(self, query: str) -> NamedMapping[Model]:
        matching_models = {
            model["model_id"]: Model.from_dict(data=model, client=self.__client)
            for model in self.__client.search.search_models_for_project(
                app_id=self.__app_id, query=query
            )
        }

        return NamedMapping(matching_models)


class ModelVersion(BaseModel):
    app_id: str = Field(validation_alias=AliasChoices("app_id", "project_id"))
    model_id: str
    version_id: str
    version: str
    hyperparameters: dict[str, Any]
    metrics: dict[str, Any]
    created_at: datetime
    modified_at: datetime
    __client: Client = PrivateAttr()

    @property
    def name(self) -> str:
        # Implement the named protocol
        return self.version

    @classmethod
    def from_dict(cls, app_id: str, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate({"app_id": app_id, **data})
        self.__client = client
        return self

    def describe(self) -> Mapping[str, Any]:
        return self.__client.component.get_model_version(
            app_id=self.app_id, version_id=self.version_id
        )


class ModelDirectoryBuilder:
    _app_id: str
    _name: str
    _parent: Directory | None
    __client: Client

    def __init__(self, client: Client, app_id: str) -> None:
        self.__client = client
        self._app_id = app_id
        self._name = ""
        self._parent = None

    def new(self, name: str) -> ModelDirectoryBuilder:
        self._name = name
        return self

    def parent(self, parent: Directory) -> ModelDirectoryBuilder:
        self._parent = parent
        return self

    def build(self) -> ModelDirectory:
        directory_id = self.__client.component.create_model_directory(
            app_id=self._app_id, name=self._name, parent=self._parent
        )
        directory_dict = self.__client.component.get_model_directory(
            app_id=self._app_id, directory_id=directory_id
        )

        return ModelDirectory.from_dict(data=directory_dict, client=self.__client)


class ModelDirectory(BaseModel):
    app_id: str = Field(validation_alias=AliasChoices("app_id", "project_id"))
    directory_id: str
    name: str
    __client: Client = PrivateAttr()

    @property
    def type(self) -> DirectoryType:
        return DirectoryType.MODEL

    @classmethod
    def from_dict(cls, data: Mapping[str, Any], client: Client) -> Self:
        logger.debug("Creating a %s from %s", cls.__name__, data)
        self = cls.model_validate(data)
        self.__client = client
        return self

    def to_dict(self) -> NamedDirectoryDict:
        return {"directory_id": self.directory_id, "type": self.type, "name": self.name}

    def directories(self) -> NamedMapping[Self]:
        directory_dicts = self.__client.component.get_model_directories_for_app(
            app_id=self.app_id, parent=self
        )
        directories = {
            directory.directory_id: directory
            for directory in (
                self.from_dict(data=directory_dict, client=self.__client)
                for directory_dict in directory_dicts
            )
        }

        return NamedMapping(directories)

    def models(self) -> NamedMapping[Model]:
        model_dicts = self.__client.component.get_models_for_app(
            app_id=self.app_id, directory_id=self.directory_id
        )

        models = {
            model.model_id: model
            for model in (
                Model.from_dict(data=model_dict, client=self.__client)
                for model_dict in model_dicts
            )
        }

        return NamedMapping(models)



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from collections import ChainMap
from collections.abc import Generator, Mapping
from itertools import chain
from typing import Any

from pydantic import (
    AliasChoices,
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator,
    model_validator,
)

from ikigai.client import datax
from ikigai.typing import Helpful
from ikigai.typing.pydantic_extensions import LowercaseStr
from ikigai.utils import CustomFacetArgumentType, FacetArgumentType
from ikigai.utils.compatibility import Self, override
from ikigai.utils.missing import MISSING, MissingType


class FacetRequirementSpec(BaseModel):
    max_child_count: int
    min_child_count: int
    max_parent_count: int
    min_parent_count: int


class ArgumentSpec(BaseModel, Helpful):
    name: str
    argument_type: FacetArgumentType
    default_value: Any | None = None
    children: dict[str, ArgumentSpec]
    have_sub_arguments: bool
    is_deprecated: bool
    is_hidden: bool
    is_list: bool
    is_required: bool
    options: list | None = None

    model_config = ConfigDict(frozen=True)

    def __validation_error_message(
        self, facet, expectation, actuals: MissingType | Any = MISSING
    ) -> str:
        if actuals is MISSING:
            actuals_str = ""
        elif actuals is None:
            actuals_str = ", got 'None'"
        elif isinstance(actuals, type):
            actuals_str = f", got type '{actuals.__name__}'"
        else:
            actuals_str = f", got {actuals.__class__.__name__}({actuals!r})"

        return f"Argument '{self.name}' for facet '{facet}' {expectation}{actuals_str}"

    def validate_value(self, facet: str, value: Any) -> None:
        if value is None:
            if self.is_required:
                error_msg = self.__validation_error_message(facet, "is required", value)
                raise ValueError(error_msg)
            return None  # No further validation for None values

        # Value is not None, perform type checking
        if self.is_list:
            return self.__validate_list_value(facet, value)

        if self.argument_type == FacetArgumentType.MAP:
            return self.__validate_dict_value(facet, value)

        # Not a dict or list, so it must be a scalar value
        return self.__validate_scalar_value(facet, value)

    def __validate_list_value(self, facet: str, value: Any) -> None:
        if not isinstance(value, list):
            error_msg = self.__validation_error_message(facet, "must be list", value)
            raise TypeError(error_msg)

        scalar_argument_spec = self.model_copy(update={"is_list": False})
        for item in value:
            scalar_argument_spec.validate_value(facet, item)
        return None  # All items validated

    def __validate_dict_value(self, facet: str, value: Any) -> None:
        if not isinstance(value, Mapping):
            error_msg = self.__validation_error_message(facet, "must be mapping", value)
            raise TypeError(error_msg)

        for name, child_value in value.items():
            if name not in self.children:
                error_msg = self.__validation_error_message(
                    facet, f"provided with unexpected child argument '{name}'"
                )
                raise KeyError(error_msg)
            child_spec = self.children[name]
            child_spec.validate_value(facet=f"{facet}:{self.name}", value=child_value)
        return None  # All child arguments validated

    def __validate_scalar_value(self, facet: str, value: Any) -> None:
        # Basic type checking based on argument_type

        if self.options and value not in self.options:
            error_msg = self.__validation_error_message(
                facet, f"must be one of {self.options}", value
            )
            raise ValueError(error_msg)

        if self.argument_type == FacetArgumentType.BOOLEAN and not isinstance(
            value, bool
        ):
            error_msg = self.__validation_error_message(facet, "must be boolean", value)
            raise TypeError(error_msg)

        if self.argument_type == FacetArgumentType.TEXT and not isinstance(value, str):
            error_msg = self.__validation_error_message(facet, "must be string", value)
            raise TypeError(error_msg)

        if self.argument_type == FacetArgumentType.NUMBER and not isinstance(
            value, int | float
        ):
            error_msg = self.__validation_error_message(facet, "must be numeric", value)
            raise TypeError(error_msg)

    @override
    def _help(self) -> Generator[str]:
        argument_type = (
            f"{self.argument_type}"
            if not self.is_list
            else f"list[{self.argument_type}]"
        )
        if self.is_required:
            argument_type += " | None"
        if not self.children:
            argument_value = f" = {self.default_value!r}" if self.default_value else ""
            yield f"{self.name}: {argument_type}{argument_value}" + (
                f"  options=[{'|'.join(self.options)}]" if self.options else ""
            )

        if self.children:
            start_brackets, end_brackets = ("[{", "}]") if self.is_list else ("{", "}")
            yield f"{self.name}: {argument_type} = " + start_brackets
            for child in self.children.values():
                if child.is_hidden:
                    continue
                yield from (f"  {child_help}" for child_help in child._help())
            yield end_brackets


class FacetInfo(BaseModel):
    facet_uid: str
    chain_group: str
    facet_group: str
    facet_type: str


class FacetType(BaseModel, Helpful):
    facet_info: FacetInfo
    is_deprecated: bool
    is_hidden: bool
    facet_requirement: FacetRequirementSpec
    facet_arguments: dict[str, ArgumentSpec]
    in_arrow_arguments: dict[str, ArgumentSpec]
    out_arrow_arguments: dict[str, ArgumentSpec]

    model_config = ConfigDict(frozen=True)

    @field_validator(
        "facet_arguments", "in_arrow_arguments", "out_arrow_arguments", mode="before"
    )
    @classmethod
    def validate_arguments(
        cls, v: list[dict] | dict[str, ArgumentSpec]
    ) -> dict[str, ArgumentSpec]:
        if isinstance(v, dict):
            return v

        if not isinstance(v, list):
            error_msg = "Expected a list of argument dictionaries"
            raise ValueError(error_msg)

        return {
            (spec := ArgumentSpec.model_validate(argument_dict)).name: spec
            for argument_dict in v
        }

    @property
    def facet_uid(self) -> str:
        return self.facet_info.facet_uid

    @property
    def name(self) -> str:
        return self.facet_info.facet_type

    @override
    def _help(self) -> Generator[str]:
        # Facet name
        yield f"{self.name.title()}:"
        # Facet Arguments
        visible_facet_arguments = [
            argument
            for argument in self.facet_arguments.values()
            if not argument.is_hidden
        ]
        if not visible_facet_arguments:
            yield "  No arguments"
            return

        if visible_facet_arguments:
            yield "  facet_arguments:"
            for argument in visible_facet_arguments:
                yield from (
                    f"    {argument_help}" for argument_help in argument._help()
                )

        # Arrow Arguments
        visible_in_arrow_arguments, out_arrow_arguments = (
            [
                argument
                for argument in self.in_arrow_arguments.values()
                if not argument.is_hidden
            ],
            [
                argument
                for argument in self.out_arrow_arguments.values()
                if not argument.is_hidden
            ],
        )

        # In Arrow Arguments
        if visible_in_arrow_arguments:
            yield "  in_arrow_arguments:"
            for argument in visible_in_arrow_arguments:
                yield from (
                    f"    {argument_help}" for argument_help in argument._help()
                )
        # Out Arrow Arguments
        if out_arrow_arguments:
            yield "  out_arrow_arguments:"
            for argument in out_arrow_arguments:
                yield from (
                    f"    {argument_help}" for argument_help in argument._help()
                )

    def is_ml_facet(self) -> bool:
        return self.facet_info.facet_group.upper() == "MACHINE_LEARNING"


class FacetTypes(BaseModel, Helpful):
    class ChainGroup(RootModel, Helpful):
        root: dict[LowercaseStr, FacetType]

        @model_validator(mode="after")
        def validate_lowercase_keys(self) -> Self:
            self.root = {
                key.lower().replace("_", "").replace(" ", ""): value
                for key, value in self.root.items()
            }
            return self

        def __contains__(self, name: str) -> bool:
            key = name.lower().replace("_", "").replace(" ", "")
            return key in self.root

        def __getitem__(self, name: str) -> FacetType:
            if name not in self:
                error_msg = f"{name.title()} facet does not exist"
                raise AttributeError(error_msg)
            key = name.lower().replace("_", "").replace(" ", "")
            return self.root[key]

        def __getattr__(self, name: str) -> FacetType:
            return self[name]

        @property
        def types(self) -> list[str]:
            return [
                facet_type.name
                for facet_type in self.root.values()
                if not facet_type.is_hidden
            ]

        def __repr__(self) -> str:
            keys = list(self.root.keys())
            return f"ChainGroup({keys})"

        def __dir__(self) -> list[str]:
            # Default dir() will return the attributes of the class
            attributes = list(super().__dir__())

            # Add the keys from the chain group
            attributes.extend([key.upper() for key in self.root])
            return attributes

        @override
        def _help(self) -> Generator[str]:
            for facet_spec in self.root.values():
                yield from (f"  {facet_help}" for facet_help in facet_spec._help())

    INPUT: ChainGroup
    MID: ChainGroup
    OUTPUT: ChainGroup

    model_config = ConfigDict(frozen=True)

    @classmethod
    def from_dict(cls, data: datax.FacetSpecsDict) -> Self:
        flattened_data = {
            "INPUT": ChainMap(*data["INPUT"].values()),
            "MID": ChainMap(*data["MID"].values()),
            "OUTPUT": ChainMap(*data["OUTPUT"].values()),
        }

        return cls.model_validate(flattened_data)

    def find_by_uid(self, uid: str) -> FacetType:
        facet_iterator = chain(
            self.INPUT.root.values(), self.MID.root.values(), self.OUTPUT.root.values()
        )
        for facet_type in facet_iterator:
            if facet_type.facet_info.facet_uid == uid:
                return facet_type
        error_msg = f"Facet type with UID {uid} not found"
        raise KeyError(error_msg)

    @override
    def _help(self) -> Generator[str]:
        # INPUT Chain
        yield "INPUT"
        yield from (f"  {chain_help}" for chain_help in self.INPUT._help())
        # MID Chain
        yield "MID"
        yield from (f"  {chain_help}" for chain_help in self.MID._help())
        # OUTPUT Chain
        yield "OUTPUT"
        yield from (f"  {chain_help}" for chain_help in self.OUTPUT._help())


class CustomFacetArgumentSpec(BaseModel):
    name: str
    argument_type: CustomFacetArgumentType = Field(
        validation_alias=AliasChoices("type", "argument_type"),
    )
    value: str | int | float | bool

    @model_validator(mode="after")
    def validate_value(self) -> Self:
        self.value = self.argument_type.python_type(self.value)
        return self

    def to_dict(self) -> datax.CustomFacetArgumentDict:
        return {
            "name": self.name,
            "type": self.argument_type.value,
            "value": str(self.value),
        }


class CustomFacetType(FacetType):
    custom_facet_id: str
    version_id: str
    custom_facet_arguments: dict[str, ArgumentSpec]

    @classmethod
    def from_facet_type(
        cls,
        facet_type: FacetType,
        custom_facet_id: str,
        version_id: str,
        custom_facet_argument_specs: dict[str, CustomFacetArgumentSpec],
    ) -> Self:
        return cls.model_validate(
            {
                "facet_info": facet_type.facet_info,
                "is_deprecated": facet_type.is_deprecated,
                "is_hidden": facet_type.is_hidden,
                "facet_requirement": facet_type.facet_requirement,
                "facet_arguments": {
                    **facet_type.facet_arguments,
                    "custom_facet_id": (
                        facet_type.facet_arguments["custom_facet_id"].model_copy(
                            update={"options": [custom_facet_id]}
                        )
                    ),
                    "version_id": (
                        facet_type.facet_arguments["version_id"].model_copy(
                            update={"options": [version_id]}
                        )
                    ),
                },
                "custom_facet_id": custom_facet_id,
                "version_id": version_id,
                "custom_facet_arguments": {
                    name: ArgumentSpec(
                        name=name,
                        argument_type=argument_spec.argument_type.to_facet_argument_type(),
                        default_value=argument_spec.value,
                        children={},
                        have_sub_arguments=False,
                        is_deprecated=False,
                        is_hidden=False,
                        is_list=False,
                        is_required=True,
                        options=None,
                    )
                    for name, argument_spec in custom_facet_argument_specs.items()
                },
                "in_arrow_arguments": facet_type.in_arrow_arguments,
                "out_arrow_arguments": facet_type.out_arrow_arguments,
            }
        )



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT


from contextlib import ExitStack
from pathlib import Path

import pytest

from ikigai import FlowStatus, Ikigai


def test_custom_facet_creation_1(
    ikigai: Ikigai,
    custom_facet_name: str,
    cleanup: ExitStack,
) -> None:
    with pytest.raises(RuntimeError, match="Custom facet was not found"):
        ikigai.custom_facets[custom_facet_name]

    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.INPUT.CUSTOM_FACET
        )
        .script(script="result = data  # no-op", arguments={"input": "input-1"})
        .build()
    )
    cleanup.callback(custom_facet.delete)

    custom_facet = ikigai.custom_facets[custom_facet_name]
    assert custom_facet is not None
    assert custom_facet.name == custom_facet_name
    assert custom_facet.facet_type == facet_types.INPUT.CUSTOM_FACET
    assert custom_facet.arguments
    assert len(custom_facet.arguments) == 1
    assert "input" in custom_facet.arguments
    input_argument = custom_facet.arguments["input"]
    assert input_argument.name == "input"
    assert input_argument.argument_type == "str"
    assert input_argument.value == "input-1"


def test_custom_facet_creation_2(
    ikigai: Ikigai,
    custom_facet_name: str,
    tmp_path: Path,
    cleanup: ExitStack,
) -> None:
    requirements_file = tmp_path / "requirements.txt"
    requirements_file.write_text("ikigai\npandas==1.5.2\n")

    script_file = tmp_path / "script.py"
    script_file.write_text(
        """
        import ikigai

        result = data  # no-op
        """
    )

    # Also test system_access flag but only for pre-setup user
    system_access = ikigai.user_email == "harsh+github-ci@ikigailabs.io"

    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.MID.CUSTOM_FACET
        )
        .script(
            script_file, requirements=requirements_file, system_access=system_access
        )
        .description("A test custom facet")
        .build()
    )
    cleanup.callback(custom_facet.delete)

    assert custom_facet.name == custom_facet_name


def test_custom_facet_creation_browsing(
    ikigai: Ikigai,
    custom_facet_name: str,
    custom_facet_name_2: str,
    custom_facet_name_3: str,
    cleanup: ExitStack,
) -> None:
    facet_types = ikigai.facet_types
    custom_facet_1 = ikigai.custom_facet.new(
        name=custom_facet_name, facet_type=facet_types.MID.CUSTOM_FACET
    ).build()
    cleanup.callback(custom_facet_1.delete)

    custom_facet_2 = (
        ikigai.custom_facet.new(
            name=custom_facet_name_2, facet_type=facet_types.MID.CUSTOM_FACET
        )
        .script(script="result = data  # no-op")
        .build()
    )
    cleanup.callback(custom_facet_2.delete)

    custom_facet_3 = (
        ikigai.custom_facet.new(
            name=custom_facet_name_3, facet_type=facet_types.MID.CUSTOM_FACET
        )
        .script(script="result = data  # no-op")
        .build()
    )
    cleanup.callback(custom_facet_3.delete)

    assert custom_facet_1 in ikigai.custom_facets.search(custom_facet_name)
    assert custom_facet_2 in ikigai.custom_facets.search(custom_facet_name_2)
    assert custom_facet_3 in ikigai.custom_facets.search(custom_facet_name_3)


def test_custom_facet_editing(
    ikigai: Ikigai,
    custom_facet_name: str,
    cleanup: ExitStack,
) -> None:
    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.MID.CUSTOM_FACET
        )
        .script(script="result = data  # no-op", arguments={"input": "input-1"})
        .description("A custom facet for testing editing operations")
        .build()
    )
    cleanup.callback(custom_facet.delete)

    (
        custom_facet.rename(name=f"updated-{custom_facet_name}")
        .update_description(description="An updated test custom facet")
        .update_script(
            script="result = data  # no-op",
            arguments={"updated-input": "input-2"},
        )
    )

    # Check the custom facet after editing -- we should get the same custom facet back
    # and the values are correctly updated
    custom_facet_after_edit = ikigai.custom_facets[f"updated-{custom_facet_name}"]
    assert custom_facet_after_edit.custom_facet_id == custom_facet.custom_facet_id
    assert (
        custom_facet_after_edit.name
        == custom_facet.name
        == f"updated-{custom_facet_name}"
    )
    assert (
        custom_facet_after_edit.description
        == custom_facet.description
        == "An updated test custom facet"
    )
    assert (
        custom_facet_after_edit.facet_type
        == custom_facet.facet_type
        == facet_types.MID.CUSTOM_FACET
    )

    assert custom_facet_after_edit.arguments == custom_facet.arguments
    assert len(custom_facet_after_edit.arguments) == 1
    assert "updated-input" in custom_facet_after_edit.arguments
    assert custom_facet_after_edit.arguments["updated-input"].value == "input-2"


def test_custom_facet_unpinned_version_run(
    ikigai: Ikigai,
    custom_facet_name: str,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test custom facet unpinned version run")
        .build()
    )
    cleanup.callback(app.delete)

    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.INPUT.CUSTOM_FACET
        )
        .script(
            script="""
            import numpy as np
            import pandas as pd

            # Create a dummy dataframe
            df = pd.DataFrame({
                "col1": np.arange(10),
                "col2": [my_input] * 10,
            })

            # Output the dummy dataframe
            result = df
            """,
            arguments={"my_input": 1},
        )
        .build()
    )
    cleanup.callback(custom_facet.delete)

    # Create a flow with the custom facet
    flow_definition = (
        ikigai.builder.custom_facet(custom_facet_version=custom_facet.unpinned())
        .arguments(my_input=2)
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(dataset_name=f"output-{flow_name}", file_type="csv", header=True)
        .build()
    )
    flow = app.flow.new(name=flow_name).definition(definition=flow_definition).build()

    # Run the flow
    log = flow.run()
    assert log.status == FlowStatus.SUCCESS, log.data
    assert log.erroneous_facet_id is None, log
    assert not log.data

    # Check the output dataset
    output_dataset = app.datasets[f"output-{flow_name}"]
    assert output_dataset is not None
    assert output_dataset.name == f"output-{flow_name}"

    output_data = output_dataset.df()
    assert output_data is not None
    assert len(output_data) > 1
    assert output_data["col1"].tolist() == list(range(10))
    assert output_data["col2"].tolist() == [2] * 10


def test_custom_facet_version_creation(
    ikigai: Ikigai,
    custom_facet_name: str,
    app_name: str,
    flow_name: str,
    cleanup: ExitStack,
) -> None:
    app = (
        ikigai.app.new(name=app_name)
        .description("App to test custom facet version creation")
        .build()
    )
    cleanup.callback(app.delete)

    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.INPUT.CUSTOM_FACET
        )
        .script(script="data = data  # bad-script")
        .build()
    )
    cleanup.callback(custom_facet.delete)

    unpinned_version = custom_facet.unpinned()

    # Create a flow with the unpinned version, try running it to see if it fails
    flow_definition_unpinned = (
        ikigai.builder.custom_facet(custom_facet_version=unpinned_version)
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(
            dataset_name=f"output-{flow_name}-unpinned", file_type="csv", header=True
        )
        .build()
    )
    flow_unpinned = (
        app.flow.new(name=f"{flow_name}-unpinned")
        .definition(definition=flow_definition_unpinned)
        .build()
    )
    cleanup.callback(flow_unpinned.delete)

    log_unpinned_1 = flow_unpinned.run()
    assert log_unpinned_1.status == FlowStatus.FAILED, log_unpinned_1.data
    assert log_unpinned_1.erroneous_facet_id, log_unpinned_1
    assert log_unpinned_1.data, log_unpinned_1

    # Create a version 1 of the custom facet
    version_1 = custom_facet.create_version(name="version-1")

    # Create a flow with the version 1, try running it to see if it fails
    flow_definition_1 = (
        ikigai.builder.custom_facet(custom_facet_version=version_1)
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(dataset_name=f"output-{flow_name}-1", file_type="csv", header=True)
        .build()
    )
    flow_1 = (
        app.flow.new(name=f"{flow_name}-1")
        .definition(definition=flow_definition_1)
        .build()
    )

    log_1 = flow_1.run()
    assert log_1.status == FlowStatus.FAILED, log_1.data
    assert log_1.erroneous_facet_id, log_1
    assert log_1.data, log_1

    # Unpinned version should still fail
    log_unpinned_2 = flow_unpinned.run()
    assert log_unpinned_2.status == FlowStatus.FAILED, log_unpinned_2.data
    assert log_unpinned_2.erroneous_facet_id, log_unpinned_2
    assert log_unpinned_2.data, log_unpinned_2

    # Update the custom facet script to fix the error
    custom_facet.update_script(
        script="""
        import numpy as np
        import pandas as pd

        # Create a dummy dataframe
        df = pd.DataFrame({
            "col1": np.arange(10),
            "col2": [1] * 10,
        })

        # Output the dummy dataframe
        result = df
        """
    )

    # Unpinned version should still fail since latest version is still bad
    log_unpinned_3 = flow_unpinned.run()
    assert log_unpinned_3.status == FlowStatus.FAILED, log_unpinned_3.data
    assert log_unpinned_3.erroneous_facet_id, log_unpinned_3
    assert log_unpinned_3.data, log_unpinned_3

    # Create a version 2 of the custom facet
    version_2 = custom_facet.create_version(name="version-2")

    # Create a flow with the version 2, try running it to see if it succeeds
    flow_definition_2 = (
        ikigai.builder.custom_facet(custom_facet_version=version_2)
        .facet(facet_type=facet_types.OUTPUT.EXPORTED, name="output")
        .arguments(dataset_name=f"output-{flow_name}-2", file_type="csv", header=True)
        .build()
    )
    flow_2 = (
        app.flow.new(name=f"{flow_name}-2")
        .definition(definition=flow_definition_2)
        .build()
    )

    log_2 = flow_2.run()
    assert log_2.status == FlowStatus.SUCCESS, log_2.data
    assert log_2.erroneous_facet_id is None, log_2
    assert not log_2.data

    # Unpinned version should now succeed
    log_unpinned_4 = flow_unpinned.run()
    assert log_unpinned_4.status == FlowStatus.SUCCESS, log_unpinned_4.data
    assert log_unpinned_4.erroneous_facet_id is None, log_unpinned_4
    assert not log_unpinned_4.data


def test_custom_facet_version_browsing(
    ikigai: Ikigai,
    custom_facet_name: str,
    custom_facet_version_name_1: str,
    custom_facet_version_name_2: str,
    custom_facet_version_name_3: str,
    cleanup: ExitStack,
) -> None:
    facet_types = ikigai.facet_types
    custom_facet = (
        ikigai.custom_facet.new(
            name=custom_facet_name, facet_type=facet_types.MID.CUSTOM_FACET
        )
        .script(script="result = data  # no-op")
        .build()
    )
    cleanup.callback(custom_facet.delete)

    created_versions = {
        custom_facet_version_name_1: custom_facet.create_version(
            name=custom_facet_version_name_1
        ),
        custom_facet_version_name_2: custom_facet.create_version(
            name=custom_facet_version_name_2
        ),
        custom_facet_version_name_3: custom_facet.create_version(
            name=custom_facet_version_name_3
        ),
    }

    versions = custom_facet.versions()
    assert len(versions) == len(created_versions)

    assert all(
        created_version in versions for created_version in created_versions.values()
    )



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from typing import Any

import pytest
from pydantic import ValidationError

from ikigai import Ikigai


def test_client_init(cred: dict[str, Any]) -> None:
    ikigai = Ikigai(**cred)
    assert ikigai


def test_client_init_bad_base_url(cred: dict[str, Any], random_name: str) -> None:
    bad_cred = {
        **cred,
        "base_url": f"https://{random_name}.ikigailabs.io",
    }
    with pytest.raises(ValidationError):
        Ikigai(**bad_cred)


def test_client_apps(cred: dict[str, Any]) -> None:
    ikigai = Ikigai(**cred)
    apps = ikigai.apps()
    assert len(apps) > 0


def test_client_app_get_item(cred: dict[str, Any], random_name: str) -> None:
    ikigai = Ikigai(**cred)
    apps = ikigai.apps()
    with pytest.raises(KeyError):
        apps[random_name]


"""
Regression Testing

- Each regression test should be of the format:
    f"test_{ticket_number}_{short_desc}"
"""


@pytest.mark.skip(
    "TODO: Update test after creating app directory is available in the client"
)
def test_iplt_7641_apps(cred: dict[str, Any]) -> None:
    # TODO: update test after creating app directory is available in the client
    pass



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.typing.protocol.directory import (
    Directory,
    DirectoryDict,
    NamedDirectoryDict,
)
from ikigai.typing.protocol.model import ModelType

__all__ = [
    "Directory",
    "DirectoryDict",
    "ModelType",
    "NamedDirectoryDict",
]



# SPDX-FileCopyrightText: 2026-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Annotated

from pydantic import AwareDatetime, BeforeValidator, PlainSerializer, StringConstraints

from ikigai.typing.pydantic_extensions.custom_serializers import (
    datetime_to_timestamp,
    optional_datetime_to_timestamp,
)
from ikigai.typing.pydantic_extensions.custom_validators import (
    cron_str,
    optional_str,
)

TimestampSerializableDatetime = Annotated[
    AwareDatetime, PlainSerializer(datetime_to_timestamp)
]
TimestampSerializableOptionalDatetime = Annotated[
    AwareDatetime | None, PlainSerializer(optional_datetime_to_timestamp)
]

OptionalStr = Annotated[str | None, BeforeValidator(optional_str)]

LowercaseStr = Annotated[str, StringConstraints(to_lower=True)]

CronStr = Annotated[str, BeforeValidator(cron_str)]



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from datetime import datetime


def datetime_to_timestamp(value: datetime) -> str:
    return str(int(value.timestamp()))


def optional_datetime_to_timestamp(value: datetime | None) -> str:
    if not value:
        return ""
    return str(int(value.timestamp()))



# SPDX-FileCopyrightText: 2024-present ikigailabs.io <harsh@ikigailabs.io>
#
# SPDX-License-Identifier: MIT

from ikigai.client import datax
from ikigai.client.client import Client
from ikigai.client.session import SSLConfig

__all__ = ["Client", "SSLConfig", "datax"]
